#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial
import six

from cinderclient import exceptions as cinder_exc
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import constants
from karbor import exception
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protection_plugin
from karbor.services.protection.protection_plugins import utils
from karbor.services.protection.protection_plugins.volume \
    import volume_snapshot_plugin_schemas as volume_schemas

LOG = logging.getLogger(__name__)

volume_snapshot_opts = [
    cfg.IntOpt(
        'poll_interval', default=15,
        help='Poll interval for Cinder volume status.'
    )
]

VOLUME_FAILURE_STATUSES = {'error', 'error_deleting', 'deleting',
                           'not-found'}

VOLUME_IGNORE_STATUSES = {'attaching', 'creating', 'backing-up',
                          'restoring-backup'}


def get_snapshot_status(cinder_client, snapshot_id):
    return get_resource_status(cinder_client.volume_snapshots, snapshot_id,
                               'snapshot')


def get_volume_status(cinder_client, volume_id):
    return get_resource_status(cinder_client.volumes, volume_id, 'volume')


def get_resource_status(resource_manager, resource_id, resource_type):
    LOG.debug('Polling %(resource_type)s (id: %(resource_id)s)',
              {'resource_type': resource_type, 'resource_id': resource_id})
    try:
        resource = resource_manager.get(resource_id)
        status = resource.status
    except cinder_exc.NotFound:
        status = 'not-found'
    LOG.debug(
        'Polled %(resource_type)s (id: %(resource_id)s) status: %(status)s',
        {'resource_type': resource_type, 'resource_id': resource_id,
         'status': status}
    )
    return status


class ProtectOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(ProtectOperation, self).__init__()
        self._interval = poll_interval

    def _create_snapshot(self, cinder_client, volume_id, snapshot_name,
                         description, force):
        snapshot = cinder_client.volume_snapshots.create(
            volume_id=volume_id,
            name=snapshot_name,
            description=description,
            force=force
        )

        snapshot_id = snapshot.id
        is_success = utils.status_poll(
            partial(get_snapshot_status, cinder_client, snapshot_id),
            interval=self._interval,
            success_statuses={'available'},
            failure_statuses={'error', 'error_deleting', 'deleting',
                              'not-found'},
            ignore_statuses={'creating'},
            ignore_unexpected=True
        )

        if not is_success:
            try:
                snapshot = cinder_client.volume_snapshots.get(snapshot_id)
            except Exception:
                reason = 'Unable to find volume snapshot.'
            else:
                reason = 'The status of snapshot is %s' % snapshot.status
            raise exception.CreateResourceFailed(
                name="Volume Snapshot",
                reason=reason, resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE)

        return snapshot_id

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        volume_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(volume_id)
        cinder_client = ClientFactory.create_client('cinder', context)
        LOG.info('Creating volume snapshot, volume_id: %s', volume_id)
        bank_section.update_object('status',
                                   constants.RESOURCE_STATUS_PROTECTING)
        volume_info = cinder_client.volumes.get(volume_id)
        is_success = utils.status_poll(
            partial(get_volume_status, cinder_client, volume_id),
            interval=self._interval,
            success_statuses={'available', 'in-use', 'error_extending',
                              'error_restoring'},
            failure_statuses=VOLUME_FAILURE_STATUSES,
            ignore_statuses=VOLUME_IGNORE_STATUSES,
        )
        if not is_success:
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateResourceFailed(
                name="Volume Snapshot",
                reason='Volume is in a error status.',
                resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE,
            )
        resource_metadata = {
            'volume_id': volume_id,
            'size': volume_info.size
        }
        snapshot_name = parameters.get('snapshot_name', None)
        description = parameters.get('description', None)
        force = parameters.get('force', False)
        try:
            snapshot_id = self._create_snapshot(cinder_client, volume_id,
                                                snapshot_name,
                                                description, force)
        except Exception as e:
            LOG.error('Error creating snapshot (volume_id: %(volume_id)s '
                      ': %(reason)s', {'volume_id': volume_id, 'reason': e})
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateResourceFailed(
                name="Volume Snapshot",
                reason=e, resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE,
            )

        resource_metadata['snapshot_id'] = snapshot_id
        bank_section.update_object('metadata', resource_metadata)
        bank_section.update_object('status',
                                   constants.RESOURCE_STATUS_AVAILABLE)
        LOG.info('Snapshot volume (volume_id: %(volume_id)s snapshot_id: '
                 '%(snapshot_id)s ) successfully',
                 {'volume_id': volume_id, 'snapshot_id': snapshot_id})


class RestoreOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(RestoreOperation, self).__init__()
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        original_volume_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(original_volume_id)
        cinder_client = ClientFactory.create_client('cinder', context)
        resource_metadata = bank_section.get_object('metadata')
        restore_name = parameters.get('restore_name',
                                      'volume-%s@%s' % (checkpoint.id,
                                                        original_volume_id))
        restore_description = parameters.get('restore_description', None)
        snapshot_id = resource_metadata['snapshot_id']
        size = resource_metadata['size']
        restore = kwargs.get('restore')
        LOG.info("Restoring a volume from snapshot, "
                 "original_volume_id: %s", original_volume_id)
        try:
            volume = cinder_client.volumes.create(
                size, snapshot_id=snapshot_id,
                name=restore_name,
                description=restore_description)
            is_success = utils.status_poll(
                partial(get_volume_status, cinder_client, volume.id),
                interval=self._interval,
                success_statuses={'available', 'in-use', 'error_extending',
                                  'error_restoring'},
                failure_statuses=VOLUME_FAILURE_STATUSES,
                ignore_statuses=VOLUME_IGNORE_STATUSES,
            )
            if is_success is not True:
                LOG.error('The status of volume is invalid. status:%s',
                          volume.status)
                reason = 'Invalid status: %s' % volume.status
                restore.update_resource_status(
                    constants.VOLUME_RESOURCE_TYPE,
                    volume.id, volume.status, reason)
                restore.save()
                raise exception.RestoreResourceFailed(
                    name="Volume Snapshot",
                    resource_id=original_volume_id,
                    resource_type=constants.VOLUME_RESOURCE_TYPE)
            restore.update_resource_status(constants.VOLUME_RESOURCE_TYPE,
                                           volume.id, volume.status)
            restore.save()
        except Exception as e:
            LOG.error("Restore volume from snapshot failed, volume_id: %s",
                      original_volume_id)
            raise exception.RestoreResourceFailed(
                name="Volume Snapshot",
                reason=e, resource_id=original_volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE)
        LOG.info("Finish restoring a volume from snapshot, volume_id: %s",
                 original_volume_id)


class DeleteOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(DeleteOperation, self).__init__()
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        resource_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(resource_id)
        snapshot_id = None
        try:
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_DELETING)
            resource_metadata = bank_section.get_object('metadata')
            snapshot_id = resource_metadata['snapshot_id']
            cinder_client = ClientFactory.create_client('cinder', context)
            try:
                snapshot = cinder_client.volume_snapshots.get(snapshot_id)
                cinder_client.volume_snapshots.delete(snapshot)
            except cinder_exc.NotFound:
                LOG.info('Snapshot id: %s not found. Assuming deleted',
                         snapshot_id)
            is_success = utils.status_poll(
                partial(get_snapshot_status, cinder_client, snapshot_id),
                interval=self._interval,
                success_statuses={'deleted', 'not-found'},
                failure_statuses={'error', 'error_deleting'},
                ignore_statuses={'deleting'},
                ignore_unexpected=True
            )
            if not is_success:
                raise exception.NotFound()
            bank_section.delete_object('metadata')
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_DELETED)
        except Exception as e:
            LOG.error('Delete volume snapshot failed, snapshot_id: %s',
                      snapshot_id)
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.DeleteResourceFailed(
                name="Volume Snapshot",
                reason=six.text_type(e),
                resource_id=resource_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE
            )


class VolumeSnapshotProtectionPlugin(protection_plugin.ProtectionPlugin):
    _SUPPORT_RESOURCE_TYPES = [constants.VOLUME_RESOURCE_TYPE]

    def __init__(self, config=None):
        super(VolumeSnapshotProtectionPlugin, self).__init__(config)
        self._config.register_opts(volume_snapshot_opts,
                                   'volume_snapshot_plugin')
        self._plugin_config = self._config.volume_snapshot_plugin
        self._poll_interval = self._plugin_config.poll_interval

    @classmethod
    def get_supported_resources_types(cls):
        return cls._SUPPORT_RESOURCE_TYPES

    @classmethod
    def get_options_schema(cls, resources_type):
        return volume_schemas.OPTIONS_SCHEMA

    @classmethod
    def get_restore_schema(cls, resources_type):
        return volume_schemas.RESTORE_SCHEMA

    @classmethod
    def get_saved_info_schema(cls, resources_type):
        return volume_schemas.SAVED_INFO_SCHEMA

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        pass

    def get_protect_operation(self, resource):
        return ProtectOperation(self._poll_interval)

    def get_restore_operation(self, resource):
        return RestoreOperation(self._poll_interval)

    def get_delete_operation(self, resource):
        return DeleteOperation(self._poll_interval)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Operation(object):
    def on_prepare_begin(self, checkpoint, resource, context, parameters,
                         **kwargs):
        """on_prepare_begin hook runs before any child resource's hooks run

        Optional
        :param checkpoint: checkpoint object for this operation
        :param resource: a resource object for this operation
        :param context: current operation context (viable for clients)
        :param parameters: dictionary representing operation parameters
        :param restore: Restore object for restore operation only
        :param heat_template: HeatTemplate for restore operation only
        """
        pass

    def on_prepare_finish(self, checkpoint, resource, context, parameters,
                          **kwargs):
        """on_prepare_finish hook runs after all child resources' prepare hooks

        Optional
        :param checkpoint: checkpoint object for this operation
        :param resource: a resource object for this operation
        :param context: current operation context (viable for clients)
        :param parameters: dictionary representing operation parameters
        :param restore: Restore object for restore operation only
        :param heat_template: HeatTemplate for restore operation only
        """
        pass

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        """on_main hook runs in parallel to other resources' on_main hooks

        Your main operation heavy lifting should probably be here.
        Optional
        :param checkpoint: checkpoint object for this operation
        :param resource: a resource object for this operation
        :param context: current operation context (viable for clients)
        :param parameters: dictionary representing operation parameters
        :param restore: Restore object for restore operation only
        :param heat_template: HeatTemplate for restore operation only
        """
        pass

    def on_complete(self, checkpoint, resource, context, parameters, **kwargs):
        """on_complete hook runs after all dependent resource's hooks

        Optional
        :param checkpoint: checkpoint object for this operation
        :param resource: a resource object for this operation
        :param context: current operation context (viable for clients)
        :param parameters: dictionary representing operation parameters
        :param restore: Restore object for restore operation only
        :param heat_template: HeatTemplate for restore operation only
        """
        pass


class ProtectionPlugin(object):
    def __init__(self, config=None):
        super(ProtectionPlugin, self).__init__()
        self._config = config

    def get_protect_operation(self, resource):
        """Returns the protect Operation for this resource

        :returns: Operation for the resource
        """
        raise NotImplementedError

    def get_restore_operation(self, resource):
        """Returns the restore Operation for this resource

        :returns: Operation for the resource
        """
        raise NotImplementedError

    def get_delete_operation(self, resource):
        """Returns the delete Operation for this resource

        :returns: Operation for the resource
        """
        raise NotImplementedError

    @classmethod
    def get_supported_resources_types(cls):
        """Returns a list of resource types this plugin supports

        :returns: a list of resource types
        """
        raise NotImplementedError

    @classmethod
    def get_options_schema(cls, resource_type):
        """Returns the protect options schema for a resource type

        :returns: a dictionary representing the schema
        """
        raise NotImplementedError

    @classmethod
    def get_saved_info_schema(cls, resource_type):
        """Returns the saved info schema for a resource type

        :returns: a dictionary representing the schema
        """
        raise NotImplementedError

    @classmethod
    def get_restore_schema(cls, resource_type):
        """Returns the restore schema for a resource type

        :returns: a dictionary representing the schema
        """
        raise NotImplementedError

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        """Returns the saved info for a resource

        :returns: a dictionary representing the saved info
        """
        raise NotImplementedError



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import defaultdict
from concurrent import futures
from oslo_config import cfg
from oslo_log import log as logging
from threading import RLock

from karbor.services.operationengine.engine.executors import \
    scheduled_operation_executor as base_executor

executor_opts = [
    cfg.IntOpt('thread_count',
               default=10,
               help='The count of thread which executor will start')
]

CONF = cfg.CONF
CONF.register_opts(executor_opts)

LOG = logging.getLogger(__name__)


class ThreadPoolExecutor(base_executor.ScheduledOperationExecutor):

    def __init__(self, operation_manager, thread_count=None):
        super(ThreadPoolExecutor, self).__init__(operation_manager)

        if thread_count is None:
            thread_count = CONF.thread_count

        self._pool = futures.ThreadPoolExecutor(thread_count)
        self._operation_to_run = defaultdict(int)
        self._operation_to_cancel = set()
        self._lock = RLock()

        self._check_functions = {
            self._CHECK_ITEMS['is_waiting']: lambda op_id: (
                op_id in self._operation_to_run),

            self._CHECK_ITEMS['is_canceled']: lambda op_id: (
                op_id in self._operation_to_cancel),
        }

    def shutdown(self, wait=True):
        self._pool.shutdown(wait)
        self._operation_to_run.clear()
        self._operation_to_cancel.clear()

    def cancel_operation(self, operation_id):
        with self._lock:
            if operation_id in self._operation_to_run:
                self._operation_to_cancel.add(operation_id)

    def _check_operation(self, operation_id, check_items):
        with self._lock:
            return any(self._check_functions[item](operation_id)
                       for item in check_items)

    def _execute_operation(self, operation_id, function, param):

        def callback(f):
            self._finish_operation(operation_id)

        with self._lock:
            self._operation_to_run[operation_id] += 1

            try:
                f = self._pool.submit(function, operation_id, param)
                f.add_done_callback(callback)

            except Exception:
                self._operation_to_run[operation_id] -= 1
                LOG.exception("Submit operation(%(o_id)s) failed.",
                              operation_id)

    def _finish_operation(self, operation_id):
        with self._lock:
            self._operation_to_run[operation_id] -= 1
            if 0 == self._operation_to_run[operation_id]:
                del self._operation_to_run[operation_id]

                if operation_id in self._operation_to_cancel:
                    self._operation_to_cancel.remove(operation_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from functools import partial

from karbor.common import constants
from karbor.tests.fullstack import karbor_base as base
from karbor.tests.fullstack import utils

SHORT_TIMEOUT = 150
MEDIUM_TIMEOUT = 450
LONG_TIMEOUT = 900

SHORT_SLEEP = 3
MEDIUM_SLEEP = 15
LONG_SLEEP = 30
HUGE_SLEEP = 100

DEFAULT_FLAVOR = "cirros256"
DEFAULT_NETWORK = "private"


class Checkpoint(object):
    def __init__(self):
        super(Checkpoint, self).__init__()
        self.id = None
        self._provider_id = None
        self.karbor_client = base._get_karbor_client()

    def _checkpoint_status(self, status=None):
        try:
            cp = self.karbor_client.checkpoints.get(self._provider_id, self.id)
        except Exception:
            return False

        if status is None or cp.status == status:
            return True
        else:
            return False

    def create(self, provider_id, plan_id, timeout=LONG_TIMEOUT):
        self._provider_id = provider_id
        checkpoint = self.karbor_client.checkpoints.create(provider_id,
                                                           plan_id)
        self.id = checkpoint.id
        utils.wait_until_true(partial(self._checkpoint_status,
                                      constants.CHECKPOINT_STATUS_AVAILABLE),
                              timeout=timeout, sleep=HUGE_SLEEP)
        return self.id

    def close(self, timeout=MEDIUM_TIMEOUT):
        try:
            self.karbor_client.checkpoints.delete(self._provider_id, self.id)
        except Exception:
            return
        utils.wait_until_true(partial(self._checkpoint_status,
                                      constants.CHECKPOINT_STATUS_DELETED),
                              timeout=timeout, sleep=LONG_SLEEP)


class Plan(object):
    _name_id = 0

    def __init__(self):
        super(Plan, self).__init__()
        self.id = None
        self.karbor_client = base._get_karbor_client()

    def create(self, provider_id, resources,
               parameters={}, name=None):
        def _transform_resource(resource):
            if isinstance(resource, dict):
                return resource
            if hasattr(resource, 'to_dict') and callable(resource.to_dict):
                return resource.to_dict()

        if name is None:
            name = "KarborFullstack-Plan-{id}".format(
                id=self.__class__._name_id
            )
            self.__class__._name_id += 1

        resources = map(_transform_resource, resources)
        plan = self.karbor_client.plans.create(name, provider_id, resources,
                                               parameters)
        self.id = plan.id
        return self.id

    def update(self, data):
        return self.karbor_client.plans.update(self.id, data)

    def close(self):
        try:
            self.karbor_client.plans.delete(self.id)
        except Exception:
            return


class Restore(object):
    def __init__(self):
        super(Restore, self).__init__()
        self.id = None
        self.karbor_client = base._get_karbor_client()

    def _restore_status(self, status=None):
        try:
            restore = self.karbor_client.restores.get(self.id)
        except Exception:
            return False

        if status is None or restore.status == status:
            return True
        else:
            return False

    def create(self, provider_id, checkpoint_id, target, parameters,
               restore_auth, timeout=LONG_TIMEOUT):
        restore = self.karbor_client.restores.create(provider_id,
                                                     checkpoint_id,
                                                     target,
                                                     parameters,
                                                     restore_auth)
        self.id = restore.id
        utils.wait_until_true(partial(self._restore_status, 'success'),
                              timeout=timeout, sleep=HUGE_SLEEP)
        return self.id

    def close(self):
        pass


class Trigger(object):
    _name_id = 0

    def __init__(self):
        super(Trigger, self).__init__()
        self.id = None
        self.karbor_client = base._get_karbor_client()

    def create(self, type, properties, name=None):
        if name is None:
            name = "KarborFullstack-Trigger-{id}".format(
                id=self.__class__._name_id
            )
            self.__class__._name_id += 1

        trigger = self.karbor_client.triggers.create(name, type, properties)
        self.id = trigger.id
        return self.id

    def close(self):
        try:
            self.karbor_client.triggers.delete(self.id)
        except Exception:
            return


class ScheduledOperation(object):
    _name_id = 0

    def __init__(self):
        super(ScheduledOperation, self).__init__()
        self.id = None
        self.karbor_client = base._get_karbor_client()

    def create(self, operation_type, trigger_id,
               operation_definition, name=None):
        if name is None:
            name = "KarborFullstack-Scheduled-Operation-{id}".format(
                id=self.__class__._name_id
            )
            self.__class__._name_id += 1

        scheduled_operation = self.karbor_client.scheduled_operations.create(
            name,
            operation_type,
            trigger_id,
            operation_definition
        )
        self.id = scheduled_operation.id
        return self.id

    def close(self):
        try:
            self.karbor_client.scheduled_operations.delete(self.id)
        except Exception:
            return


class Server(object):
    _name_id = 0

    def __init__(self):
        super(Server, self).__init__()
        self.id = None
        self._name = None
        self.nova_client = base._get_nova_client()
        self.neutron_client = base._get_neutron_client()
        self.cinder_client = base._get_cinder_client()
        self.glance_client = base._get_glance_client()

    def _server_status(self, status=None):
        try:
            server = self.nova_client.servers.get(self.id)
        except Exception:
            return False

        if status is None or status == server.status:
            return True
        else:
            return False

    def to_dict(self):
        return {
            "id": self.id,
            "type": constants.SERVER_RESOURCE_TYPE,
            "name": self._name,
        }

    def create(self, name=None, image=None, volume=None, flavor=DEFAULT_FLAVOR,
               network=DEFAULT_NETWORK, timeout=LONG_TIMEOUT):
        block_device_mapping_v2 = None
        if volume:
            block_device_mapping_v2 = [{
                'uuid': volume,
                'source_type': 'volume',
                'destination_type': 'volume',
                'boot_index': 0,
                'delete_on_termination': False}]
        else:
            if not image:
                images = self.glance_client.images.list()
                for image_iter in images:
                    if image_iter['disk_format'] not in ('aki', 'ari') and (
                            image_iter['name'].startswith('cirros')):
                        image = image_iter['id']
                        break
            assert image
        flavor = self.nova_client.flavors.find(name=flavor)
        if name is None:
            name = "KarborFullstack-Server-{id}".format(
                id=self.__class__._name_id
            )
            self.__class__._name_id += 1
            self._name = name

        networks = self.neutron_client.list_networks(name=network)
        assert len(networks['networks']) > 0
        network_id = networks['networks'][0]['id']

        server = self.nova_client.servers.create(
            name=name,
            image=image,
            block_device_mapping_v2=block_device_mapping_v2,
            flavor=flavor,
            nics=[{"net-id": network_id}],
        )
        self.id = server.id

        utils.wait_until_true(partial(self._server_status, 'ACTIVE'),
                              timeout=timeout, sleep=MEDIUM_SLEEP)
        return self.id

    def _volume_attached(self, volume_id):
        volume_item = self.cinder_client.volumes.get(volume_id)
        server_attachments = filter(lambda x: x['server_id'] == self.id,
                                    volume_item.attachments)
        if len(server_attachments) > 0:
            return True
        else:
            return False

    def attach_volume(self, volume_id, timeout=MEDIUM_TIMEOUT):
        self.nova_client.volumes.create_server_volume(self.id, volume_id)
        utils.wait_until_true(partial(self._volume_attached, volume_id),
                              timeout=timeout, sleep=SHORT_SLEEP)

    def _volume_detached(self, volume_id):
        volume_item = self.cinder_client.volumes.get(volume_id)
        server_attachments = filter(lambda x: x['server_id'] == self.id,
                                    volume_item.attachments)
        if len(server_attachments) > 0:
            return False
        else:
            return True

    def detach_volume(self, volume_id, timeout=MEDIUM_TIMEOUT):
        self.nova_client.volumes.delete_server_volume(self.id, volume_id)
        utils.wait_until_true(partial(self._volume_detached, volume_id),
                              timeout=timeout, sleep=SHORT_SLEEP)

    def close(self, timeout=MEDIUM_TIMEOUT):
        try:
            self.nova_client.servers.delete(self.id)
        except Exception:
            return
        utils.wait_until_none(self._server_status, timeout=timeout,
                              sleep=MEDIUM_SLEEP)


class Volume(object):
    _name_id = 0

    def __init__(self):
        super(Volume, self).__init__()
        self.id = None
        self._name = None
        self.cinder_client = base._get_cinder_client()
        self.glance_client = base._get_glance_client()

    def _volume_status(self, status=None):
        try:
            volume = self.cinder_client.volumes.get(self.id)
        except Exception:
            return False

        if status is None or status == volume.status:
            return True
        else:
            return False

    def to_dict(self):
        return {
            "id": self.id,
            "type": constants.VOLUME_RESOURCE_TYPE,
            "name": self._name,
            "extra_info": {'availability_zone': 'az1'},
        }

    def create(self, size, name=None, create_from_image=False,
               timeout=LONG_TIMEOUT):
        if name is None:
            name = "KarborFullstack-Volume-{id}".format(
                id=self.__class__._name_id
            )
            self.__class__._name_id += 1

        self._name = name
        image = None
        if create_from_image:
            images = self.glance_client.images.list()
            for image_iter in images:
                if image_iter['disk_format'] not in ('aki', 'ari') and (
                        image_iter['name'].startswith('cirros')):
                    image = image_iter['id']
                    break
            assert image
        volume = self.cinder_client.volumes.create(size, name=name,
                                                   imageRef=image)
        self.id = volume.id
        utils.wait_until_true(partial(self._volume_status, 'available'),
                              timeout=timeout, sleep=MEDIUM_SLEEP)
        return self.id

    def close(self, timeout=LONG_TIMEOUT):
        try:
            self.cinder_client.volumes.delete(self.id)
        except Exception:
            return
        utils.wait_until_none(self._volume_status, timeout=timeout,
                              sleep=MEDIUM_SLEEP)


class Share(object):
    _name_id = 0

    def __init__(self):
        super(Share, self).__init__()
        self.id = None
        self._name = None
        self.manila_client = base._get_manila_client()
        self.neutron_client = base._get_neutron_client()

    def _share_status(self, status=None):
        try:
            share = self.manila_client.shares.get(self.id)
        except Exception:
            return False

        if status is None or status == share.status:
            return True
        else:
            return False

    def to_dict(self):
        return {
            "id": self.id,
            "type": constants.SHARE_RESOURCE_TYPE,
            "name": self._name,
        }

    def create(self, share_proto, size, name=None, timeout=LONG_TIMEOUT):
        if name is None:
            name = "KarborFullstack-Share-{id}".format(id=self._name_id)
            self._name_id += 1

        self._name = name
        share = self.manila_client.shares.create(share_proto, size, name=name)
        self.id = share.id
        utils.wait_until_true(partial(self._share_status, 'available'),
                              timeout=timeout, sleep=MEDIUM_SLEEP)
        return self.id

    def close(self, timeout=MEDIUM_TIMEOUT):
        try:
            self.manila_client.shares.delete(self.id)
        except Exception:
            return
        utils.wait_until_none(self._share_status, timeout=timeout,
                              sleep=MEDIUM_SLEEP)


class Network(object):
    def __init__(self):
        super(Network, self).__init__()
        self.id = None
        self.project_id = None
        self._name = "private-net"
        self.neutron_client = base._get_neutron_client()

    def _network_status(self, status=None):
        try:
            networks = self.neutron_client.list_networks(name=self._name)
            assert len(networks['networks']) > 0
            network = networks['networks'][0]
        except Exception:
            return False

        if status is None or status == network['status']:
            return True
        else:
            return False

    def to_dict(self):
        return {
            "id": self.id,
            "type": constants.NETWORK_RESOURCE_TYPE,
            "name": self._name,
        }

    def create(self, timeout=MEDIUM_TIMEOUT):
        network = {'name': self._name, 'admin_state_up': True}
        self.neutron_client.create_network({'network': network})

        networks = self.neutron_client.list_networks(name=self._name)
        assert len(networks['networks']) > 0
        network_id = networks['networks'][0]['id']
        self.id = network_id
        self.project_id = networks['networks'][0]['tenant_id']

        utils.wait_until_true(partial(self._network_status, 'ACTIVE'),
                              timeout=timeout, sleep=MEDIUM_SLEEP)

        return self.id

    def close(self, timeout=LONG_TIMEOUT):
        try:
            self.neutron_client.delete_network(self.id)
        except Exception:
            return
        utils.wait_until_none(self._network_status, timeout=timeout,
                              sleep=MEDIUM_SLEEP)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from botocore.exceptions import ClientError


class FakeS3Client(object):
    def __init__(self, *args, **kwargs):
        super(FakeS3Client, self).__init__()

    @classmethod
    def connection(cls, *args, **kargs):
        return FakeS3Connection()


class FakeS3Connection(object):
    def __init__(self, *args, **kwargs):
        super(FakeS3Connection, self).__init__()
        self.s3_dir = {}
        self.object_headers = {}

    def create_bucket(self, Bucket):
        self.s3_dir[Bucket] = {
            'Keys': {}
        }

    def list_objects(self, Bucket, Prefix, MaxKeys, Marker):
        body = []
        prefix = '' if not Prefix else Prefix
        for obj in self.s3_dir[Bucket]['Keys'].keys():
            if obj.startswith(prefix):
                body.append({
                    'Key': obj
                })
        return {'Contents': body}

    def put_object(self, Bucket, Key, Body, Metadata=None):
        if Bucket in self.s3_dir.keys():
            self.s3_dir[Bucket]['Keys'][Key] = {
                'Body': FakeS3Stream(Body),
                'Metadata': Metadata if Metadata else {}
            }
        else:
            raise ClientError("error_bucket")

    def get_object(self, Bucket, Key):
        if Bucket in self.s3_dir.keys():
            if Key in self.s3_dir[Bucket]['Keys'].keys():
                return self.s3_dir[Bucket]['Keys'][Key]
            else:
                raise ClientError("error_object")
        else:
            raise ClientError("error_bucket")

    def delete_object(self, Bucket, Key):
        if Bucket in self.s3_dir.keys():
            if Key in self.s3_dir[Bucket]['Keys'].keys():
                del self.s3_dir[Bucket]['Keys'][Key]
            else:
                raise ClientError("error_object")
        else:
            raise ClientError("error_bucket")


class FakeS3Stream(object):
    def __init__(self, data):
        self.data = data

    def read(self):
        return self.data



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import eventlet

from datetime import datetime
from functools import partial

from karbor.common import constants
from karbor.services.operationengine.engine.triggers.timetrigger \
    .timeformats import calendar_time
from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects
from karbor.tests.fullstack import utils

pattern = "BEGIN:VEVENT\nRRULE:FREQ=WEEKLY;INTERVAL=1;\nEND:VEVENT"
DEFAULT_PROPERTY = {'pattern': pattern}


class ScheduledOperationsTest(karbor_base.KarborBaseTest):
    """Test Scheduled Operations operation

    """
    def setUp(self):
        super(ScheduledOperationsTest, self).setUp()
        providers = self.provider_list()
        self.assertTrue(len(providers))
        self.provider_id = self.provider_id_noop

    def _create_scheduled_operation(
            self,
            resources,
            trigger_properties=DEFAULT_PROPERTY,
            operation_name=None):
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, resources)
        operation_definition = {'plan_id': plan.id,
                                'provider_id': self.provider_id}
        trigger = self.store(objects.Trigger())
        trigger.create('time', trigger_properties)

        operation = objects.ScheduledOperation()
        operation.create('protect', trigger.id,
                         operation_definition, operation_name)
        return operation

    def _create_for_volume(self,
                           trigger_properties=DEFAULT_PROPERTY,
                           operation_name=None):
        volume = self.store(objects.Volume())
        volume.create(1)
        return self._create_scheduled_operation([volume, ],
                                                trigger_properties,
                                                operation_name)

    def _create_for_server(self,
                           trigger_properties=DEFAULT_PROPERTY,
                           operation_name=None):
        server = self.store(objects.Server())
        server.create()
        return self._create_scheduled_operation([server, ],
                                                trigger_properties,
                                                operation_name)

    def test_scheduled_operations_create_no_scheduled(self):
        name = "KarborFullstack-Scheduled-Operation-no-scheduled"
        operation = self.store(self._create_for_volume(operation_name=name))

        item = self.karbor_client.scheduled_operations.get(operation.id)
        self.assertEqual(name, item.name)

        items = self.karbor_client.scheduled_operations.list()
        ids = [item_.id for item_ in items]
        self.assertTrue(operation.id in ids)

    @staticmethod
    def _wait_timestamp(pattern, start_time, freq):
        if not isinstance(freq, int) or freq <= 0:
            return 0

        cur_time = copy.deepcopy(start_time)
        cal_obj = calendar_time.ICal(start_time, pattern)
        for i in range(freq):
            next_time = cal_obj.compute_next_time(cur_time)
            cur_time = next_time
        return (next_time - start_time).seconds

    def _checkpoint_status(self, checkpoint_id, status):
        try:
            cp = self.karbor_client.checkpoints.get(self.provider_id,
                                                    checkpoint_id)
        except Exception:
            return False

        if status is None or cp.status == status:
            return True
        else:
            return False

    def test_scheduled_operations_create_and_scheduled(self):
        freq = 2
        pattern = "BEGIN:VEVENT\nRRULE:FREQ=MINUTELY;INTERVAL=5;\nEND:VEVENT"
        cur_property = {'pattern': pattern, 'format': 'calendar'}

        operation = self.store(self._create_for_volume(cur_property))
        start_time = datetime.now().replace(microsecond=0)
        sleep_time = self._wait_timestamp(pattern, start_time, freq)
        self.assertNotEqual(0, sleep_time)
        eventlet.sleep(sleep_time)

        items = self.karbor_client.checkpoints.list(self.provider_id)
        operation_item = self.karbor_client.scheduled_operations.get(
            operation.id)
        plan_id = operation_item.operation_definition["plan_id"]
        cps = filter(lambda x: x.protection_plan["id"] == plan_id, items)
        self.assertEqual(freq, len(cps))

        for cp in cps:
            utils.wait_until_true(
                partial(self._checkpoint_status,
                        cp.id,
                        constants.CHECKPOINT_STATUS_AVAILABLE),
                timeout=objects.LONG_TIMEOUT, sleep=objects.LONG_SLEEP
            )
            checkpoint = self.store(objects.Checkpoint())
            checkpoint._provider_id = self.provider_id
            checkpoint.id = cp.id

    def test_scheduled_operations_list(self):
        operation1 = self.store(self._create_for_volume())
        operation2 = self.store(self._create_for_server())

        items = self.karbor_client.scheduled_operations.list()
        ids = [item.id for item in items]
        self.assertTrue(operation1.id in ids)
        self.assertTrue(operation2.id in ids)

    def test_scheduled_operations_get(self):
        name = "KarborFullstack-Scheduled-Operation-Test-Get"
        operation = self._create_for_volume(operation_name=name)
        self.store(operation)

        item = self.karbor_client.scheduled_operations.get(operation.id)
        self.assertEqual(item.name, name)
        self.assertEqual(item.id, operation.id)

    def test_scheduled_operations_delete(self):
        name = "KarborFullstack-Scheduled-Operation-Test-Delete"
        operation = self._create_for_volume(operation_name=name)

        item = self.karbor_client.scheduled_operations.get(operation.id)
        self.assertEqual(name, item.name)

        operation.close()
        items = self.karbor_client.scheduled_operations.list()
        ids = [item_.id for item_ in items]
        self.assertTrue(operation.id not in ids)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.common.constants import RESOURCE_TYPES
from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects


class ProtectablesTest(karbor_base.KarborBaseTest):
    """Test Protectables operation """

    def test_protectables_list(self):
        items = self.karbor_client.protectables.list()
        query_types = [item.protectable_type for item in items]
        self.assertItemsEqual(RESOURCE_TYPES, query_types)

    def test_protectables_get(self):
        protectable_type = 'OS::Keystone::Project'
        res = self.karbor_client.protectables.get(protectable_type)
        self.assertEqual(protectable_type, res.name)

        protectable_type = 'OS::Nova::Server'
        res = self.karbor_client.protectables.get(protectable_type)
        self.assertEqual(protectable_type, res.name)

        protectable_type = 'OS::Neutron::Network'
        res = self.karbor_client.protectables.get(protectable_type)
        self.assertEqual(protectable_type, res.name)

    def test_protectables_list_instances(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        items = self.karbor_client.protectables.list_instances(
            'OS::Cinder::Volume')
        ids = [item.id for item in items]
        self.assertTrue(volume.id in ids)

        server = self.store(objects.Server())
        server.create()
        items = self.karbor_client.protectables.list_instances(
            'OS::Nova::Server')
        ids = [item.id for item in items]
        self.assertTrue(server.id in ids)

        network = self.store(objects.Network())
        network.create()
        items = self.karbor_client.protectables.list_instances(
            'OS::Neutron::Network')
        self.assertEqual(items[0].id, network.project_id)

    def test_protectables_get_instance(self):
        volume = self.store(objects.Volume())
        volume.create(1)
        instance = self.karbor_client.protectables.get_instance(
            'OS::Cinder::Volume', volume.id)
        self.assertEqual(volume.id, instance.id)

        server = self.store(objects.Server())
        server.create()
        instance = self.karbor_client.protectables.get_instance(
            'OS::Nova::Server', server.id)
        self.assertEqual(server.id, instance.id)

    def test_protectables_get_attach_volume_instance(self):
        server = self.store(objects.Server())
        server.create()

        volume = self.store(objects.Volume())
        volume.create(1)

        server.attach_volume(volume.id)
        volume_item = self.cinder_client.volumes.get(volume.id)
        ins_res = self.karbor_client.protectables.get_instance(
            'OS::Nova::Server', volume_item.attachments[0]["server_id"])
        self.assertTrue(ins_res.dependent_resources)
        self.assertEqual('OS::Cinder::Volume',
                         ins_res.dependent_resources[0]["type"])
        self.assertEqual(volume.id,
                         ins_res.dependent_resources[0]["id"])

    def test_share_protectables_list_instances(self):
        res_list = self.karbor_client.protectables.list_instances(
            'OS::Manila::Share')
        before_num = len(res_list)
        share = self.store(objects.Share())
        share.create("NFS", 1)
        res_list = self.karbor_client.protectables.list_instances(
            'OS::Manila::Share')
        after_num = len(res_list)
        self.assertEqual(1, after_num - before_num)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg

from karbor.common import karbor_keystone_plugin as kkp
from karbor.context import RequestContext
from karbor import exception
from karbor.services.protection.clients import utils
from karbor.tests import base


class ClientUtilsTest(base.TestCase):
    def setUp(self):
        super(ClientUtilsTest, self).setUp()

        self._service = ''
        self._public_url = 'http://127.0.0.1:8776/v3/abcd'

        service_catalog = [
            {'type': 'volumev3',
             'name': 'cinderv3',
             'endpoints': [{'publicURL': self._public_url}],
             },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    @mock.patch.object(cfg.ConfigOpts, '_get')
    def test_get_url_by_endpoint(self, get_opt):
        endpoint = 'http://127.0.0.1:8776/v3'
        get_opt.return_value = endpoint

        self.assertEqual(
            endpoint, utils.get_url(self._service, self._context, cfg.CONF))

        self.assertEqual(
            endpoint + '/%s' % self._context.project_id,
            utils.get_url(self._service, self._context, cfg.CONF,
                          '%(url)s/%(project)s'))

    @mock.patch.object(cfg.ConfigOpts, '_get')
    def test_get_url_by_catalog(self, get_opt):
        def _get_opt(name):
            if name.find('catalog_info') >= 0:
                return 'volumev3:cinderv3:publicURL'
            return None
        get_opt.side_effect = _get_opt

        self.assertEqual(
            self._public_url,
            utils.get_url(self._service, self._context, cfg.CONF))

    @mock.patch.object(kkp.KarborKeystonePlugin, 'get_service_endpoint')
    def test_get_url_by_keystone_plugin(self, get_endpoint):
        endpoint = "http://127.0.0.1:8776"
        keystone_plugin = kkp.KarborKeystonePlugin()
        get_endpoint.return_value = endpoint

        config = mock.Mock()
        config.test_service_endpoint = None
        config.test_service_catalog_info = None
        self.assertEqual(
            endpoint,
            utils.get_url('test_service', self._context, config,
                          keystone_plugin=keystone_plugin))

    @mock.patch.object(cfg.ConfigOpts, '_get')
    def test_get_url_except(self, get_opt):
        get_opt.return_value = None
        self.assertRaisesRegex(exception.KarborException,
                               "Couldn't find the endpoint of service.*",
                               utils.get_url, self._service,
                               self._context, cfg.CONF)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for remote procedure calls using queue
"""

import mock
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_db import exception as db_exc

from karbor import context
from karbor import db
from karbor import exception
from karbor import manager
from karbor import rpc
from karbor import service
from karbor.tests import base


test_service_opts = [
    cfg.StrOpt("fake_manager",
               default="karbor.tests.unit.test_service.FakeManager",
               help="Manager for testing"),
    cfg.StrOpt("test_service_listen",
               help="Host to bind test service to"),
    cfg.IntOpt("test_service_listen_port",
               default=0,
               help="Port number to bind test service to"), ]

CONF = cfg.CONF
CONF.register_opts(test_service_opts)


class FakeManager(manager.Manager):
    """Fake manager for tests."""
    def __init__(self, host=None,
                 db_driver=None, service_name=None):
        super(FakeManager, self).__init__(host=host,
                                          db_driver=db_driver)

    def test_method(self):
        return 'manager'


class ExtendedService(service.Service):
    def test_method(self):
        return 'service'


class ServiceManagerTestCase(base.TestCase):
    """Test cases for Services."""

    def test_message_gets_to_manager(self):
        serv = service.Service('test',
                               'test',
                               'test',
                               'karbor.tests.unit.test_service.FakeManager')
        serv.start()
        self.assertEqual('manager', serv.test_method())

    def test_override_manager_method(self):
        serv = ExtendedService('test',
                               'test',
                               'test',
                               'karbor.tests.unit.test_service.FakeManager')
        serv.start()
        self.assertEqual('service', serv.test_method())


class ServiceFlagsTestCase(base.TestCase):
    def test_service_enabled_on_create_based_on_flag(self):
        self.flags(enable_new_services=True)
        host = 'foo'
        binary = 'karbor-fake'
        app = service.Service.create(host=host, binary=binary)
        app.start()
        app.stop()
        ref = db.service_get(context.get_admin_context(), app.service_id)
        db.service_destroy(context.get_admin_context(), app.service_id)
        self.assertFalse(ref['disabled'])

    def test_service_disabled_on_create_based_on_flag(self):
        self.flags(enable_new_services=False)
        host = 'foo'
        binary = 'karbor-fake'
        app = service.Service.create(host=host, binary=binary)
        app.start()
        app.stop()
        ref = db.service_get(context.get_admin_context(), app.service_id)
        db.service_destroy(context.get_admin_context(), app.service_id)
        self.assertTrue(ref['disabled'])


class ServiceTestCase(base.TestCase):
    """Test cases for Services."""

    def setUp(self):
        super(ServiceTestCase, self).setUp()
        self.host = 'foo'
        self.binary = 'karbor-fake'
        self.topic = 'fake'

    def test_create(self):
        app = service.Service.create(host=self.host,
                                     binary=self.binary,
                                     topic=self.topic)

        self.assertTrue(app)

    def test_report_state_newly_disconnected(self):
        service_ref = {'host': self.host,
                       'binary': self.binary,
                       'topic': self.topic,
                       'report_count': 0,
                       'id': 1}
        with mock.patch.object(service, 'db') as mock_db:
            mock_db.service_get_by_args.side_effect = exception.NotFound()
            mock_db.service_create.return_value = service_ref
            mock_db.service_get.side_effect = db_exc.DBConnectionError()

            serv = service.Service(
                self.host,
                self.binary,
                self.topic,
                'karbor.tests.unit.test_service.FakeManager'
            )
            serv.start()
            serv.report_state()
            self.assertTrue(serv.model_disconnected)
            self.assertFalse(mock_db.service_update.called)

    def test_report_state_disconnected_DBError(self):
        service_ref = {'host': self.host,
                       'binary': self.binary,
                       'topic': self.topic,
                       'report_count': 0,
                       'id': 1}
        with mock.patch.object(service, 'db') as mock_db:
            mock_db.service_get_by_args.side_effect = exception.NotFound()
            mock_db.service_create.return_value = service_ref
            mock_db.service_get.side_effect = db_exc.DBError()

            serv = service.Service(
                self.host,
                self.binary,
                self.topic,
                'karbor.tests.unit.test_service.FakeManager'
            )
            serv.start()
            serv.report_state()
            self.assertTrue(serv.model_disconnected)
            self.assertFalse(mock_db.service_update.called)

    def test_report_state_newly_connected(self):
        service_ref = {'host': self.host,
                       'binary': self.binary,
                       'topic': self.topic,
                       'report_count': 0,
                       'id': 1}
        with mock.patch.object(service, 'db') as mock_db:
            mock_db.service_get_by_args.side_effect = exception.NotFound()
            mock_db.service_create.return_value = service_ref
            mock_db.service_get.return_value = service_ref

            serv = service.Service(
                self.host,
                self.binary,
                self.topic,
                'karbor.tests.unit.test_service.FakeManager'
            )
            serv.start()
            serv.model_disconnected = True
            serv.report_state()

            self.assertFalse(serv.model_disconnected)
            self.assertTrue(mock_db.service_update.called)

    def test_report_state_manager_not_working(self):
        service_ref = {'host': self.host,
                       'binary': self.binary,
                       'topic': self.topic,
                       'report_count': 0,
                       'id': 1}
        with mock.patch('karbor.db') as mock_db:
            mock_db.service_get.return_value = service_ref

            serv = service.Service(
                self.host,
                self.binary,
                self.topic,
                'karbor.tests.unit.test_service.FakeManager'
            )
            serv.manager.is_working = mock.Mock(return_value=False)
            serv.start()
            serv.report_state()

            serv.manager.is_working.assert_called_once_with()
            self.assertFalse(mock_db.service_update.called)

    def test_service_with_long_report_interval(self):
        self.override_config('service_down_time', 10)
        self.override_config('report_interval', 10)
        service.Service.create(
            binary="test_service",
            manager="karbor.tests.unit.test_service.FakeManager")
        self.assertEqual(25, CONF.service_down_time)

    @mock.patch.object(rpc, 'get_server')
    @mock.patch('karbor.db')
    def test_service_stop_waits_for_rpcserver(self, mock_db, mock_rpc):
        serv = service.Service(
            self.host,
            self.binary,
            self.topic,
            'karbor.tests.unit.test_service.FakeManager'
        )
        serv.start()
        serv.stop()
        serv.wait()
        serv.rpcserver.start.assert_called_once_with()
        serv.rpcserver.stop.assert_called_once_with()
        serv.rpcserver.wait.assert_called_once_with()


class TestWSGIService(base.TestCase):

    def setUp(self):
        super(TestWSGIService, self).setUp()

    @mock.patch('oslo_service.wsgi.Loader')
    def test_service_random_port(self, mock_loader):
        test_service = service.WSGIService("test_service")
        self.assertEqual(0, test_service.port)
        test_service.start()
        self.assertNotEqual(0, test_service.port)
        test_service.stop()
        self.assertTrue(mock_loader.called)

    @mock.patch('oslo_service.wsgi.Loader')
    def test_reset_pool_size_to_default(self, mock_loader):
        test_service = service.WSGIService("test_service")
        test_service.start()

        # Stopping the service, which in turn sets pool size to 0
        test_service.stop()
        self.assertEqual(0, test_service.server._pool.size)

        # Resetting pool size to default
        test_service.reset()
        test_service.start()
        self.assertEqual(cfg.CONF.wsgi_default_pool_size,
                         test_service.server._pool.size)
        self.assertTrue(mock_loader.called)

    @mock.patch('oslo_service.wsgi.Loader')
    def test_workers_set_default(self, mock_loader):
        self.override_config('osapi_karbor_listen_port',
                             CONF.test_service_listen_port)
        test_service = service.WSGIService("osapi_karbor")
        self.assertEqual(processutils.get_worker_count(),
                         test_service.workers)
        self.assertTrue(mock_loader.called)

    @mock.patch('oslo_service.wsgi.Loader')
    def test_workers_set_good_user_setting(self, mock_loader):
        self.override_config('osapi_karbor_listen_port',
                             CONF.test_service_listen_port)
        self.override_config('osapi_karbor_workers', 8)
        test_service = service.WSGIService("osapi_karbor")
        self.assertEqual(8, test_service.workers)
        self.assertTrue(mock_loader.called)

    @mock.patch('oslo_service.wsgi.Loader')
    def test_workers_set_zero_user_setting(self, mock_loader):
        self.override_config('osapi_karbor_listen_port',
                             CONF.test_service_listen_port)
        self.override_config('osapi_karbor_workers', 0)
        test_service = service.WSGIService("osapi_karbor")
        # If a value less than 1 is used, defaults to number of procs
        # available
        self.assertEqual(processutils.get_worker_count(),
                         test_service.workers)
        self.assertTrue(mock_loader.called)

    @mock.patch('oslo_service.wsgi.Loader')
    def test_workers_set_negative_user_setting(self, mock_loader):
        self.override_config('osapi_karbor_workers', -1)
        self.assertRaises(exception.InvalidInput,
                          service.WSGIService, "osapi_karbor")
        self.assertTrue(mock_loader.called)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile

from swiftclient import ClientException


class FakeSwiftClient(object):
    def __init__(self, *args, **kwargs):
        super(FakeSwiftClient, self).__init__()

    @classmethod
    def connection(cls, *args, **kargs):
        return FakeSwiftConnection()


class FakeSwiftConnection(object):
    def __init__(self, *args, **kwargs):
        super(FakeSwiftConnection, self).__init__()
        self.swiftdir = tempfile.mkdtemp()
        self.object_headers = {}

    def put_container(self, container):
        container_dir = self.swiftdir + "/" + container
        if os.path.exists(container_dir) is True:
            return
        else:
            os.makedirs(container_dir)

    def get_container(self, container, prefix, limit, marker, end_marker):
        container_dir = self.swiftdir + "/" + container
        body = []
        if prefix:
            objects_dir = container_dir + "/" + prefix
        else:
            objects_dir = container_dir
        for f in os.listdir(objects_dir):
            if os.path.isfile(objects_dir + "/" + f):
                body.append({"name": f})
            else:
                body.append({"subdir": f})
        return None, body

    def put_object(self, container, obj, contents, headers=None):
        container_dir = self.swiftdir + "/" + container
        obj_file = container_dir + "/" + obj
        obj_dir = obj_file[0:obj_file.rfind("/")]
        if os.path.exists(container_dir) is True:
            if os.path.exists(obj_dir) is False:
                os.makedirs(obj_dir)
            with open(obj_file, "w") as f:
                f.write(contents)

            self.object_headers[obj_file] = {}
            for key, value in headers.items():
                self.object_headers[obj_file][str(key)] = str(value)
            return
        else:
            raise ClientException("error_container")

    def get_object(self, container, obj):
        container_dir = self.swiftdir + "/" + container
        obj_file = container_dir + "/" + obj
        if os.path.exists(container_dir) is True:
            if os.path.exists(obj_file) is True:
                with open(obj_file, "r") as f:
                    return self.object_headers[obj_file], f.read()
            else:
                raise ClientException("error_obj")
        else:
            raise ClientException("error_container")

    def delete_object(self, container, obj):
        container_dir = self.swiftdir + "/" + container
        obj_file = container_dir + "/" + obj
        if os.path.exists(container_dir) is True:
            if os.path.exists(obj_file) is True:
                os.remove(obj_file)
                self.object_headers.pop(obj_file)
            else:
                raise ClientException("error_obj")
        else:
            raise ClientException("error_container")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects


class TriggersTest(karbor_base.KarborBaseTest):
    """Test Triggers operation"""

    def test_triggers_list(self):
        pattern1 = "BEGIN:VEVENT\nRRULE:FREQ=HOURLY;INTERVAL=1;\nEND:VEVENT"
        trigger1 = self.store(objects.Trigger())
        trigger1.create('time', {'pattern': pattern1, 'format': 'calendar'})
        pattern2 = "BEGIN:VEVENT\nRRULE:FREQ=WEEKLY;INTERVAL=1;\nEND:VEVENT"
        trigger2 = self.store(objects.Trigger())
        trigger2.create('time', {'pattern': pattern2, 'format': 'calendar'})

        items = self.karbor_client.triggers.list()
        ids = [item.id for item in items]
        self.assertTrue(trigger1.id in ids)
        self.assertTrue(trigger2.id in ids)

    def test_triggers_get(self):
        trigger_name = "FullStack Trigger Test Get"
        pattern = "BEGIN:VEVENT\nRRULE:FREQ=WEEKLY;INTERVAL=1;\nEND:VEVENT"
        trigger = self.store(objects.Trigger())
        trigger.create('time', {'pattern': pattern, 'format': 'calendar'},
                       name=trigger_name)
        trigger = self.karbor_client.triggers.get(trigger.id)
        self.assertEqual(trigger_name, trigger.name)

    def test_triggers_delete(self):
        pattern = "BEGIN:VEVENT\nRRULE:FREQ=WEEKLY;INTERVAL=1;\nEND:VEVENT"
        trigger = objects.Trigger()
        trigger.create('time', {'pattern': pattern, 'format': 'calendar'})
        self.karbor_client.triggers.delete(trigger.id)
        items = self.karbor_client.triggers.list()
        ids = [item.id for item in items]
        self.assertTrue(trigger.id not in ids)



# Copyright (C) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for db purge."""

import datetime

from oslo_db import exception as db_exc
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy.dialects import sqlite

from karbor import context
from karbor import db
from karbor.db.sqlalchemy import api as db_api
from karbor import exception
from karbor.tests import base

from oslo_db.sqlalchemy import utils as sqlalchemyutils


class PurgeDeletedTest(base.TestCase):

    def setUp(self):
        super(PurgeDeletedTest, self).setUp()
        self.context = context.get_admin_context()
        self.engine = db_api.get_engine()
        self.session = db_api.get_session()
        self.conn = self.engine.connect()
        self.plans = sqlalchemyutils.get_table(
            self.engine, "plans")
        # The resources table has a FK of plans.id
        self.resources = sqlalchemyutils.get_table(
            self.engine, "resources")

        self.uuidstrs = []
        for unused in range(6):
            self.uuidstrs.append(uuidutils.generate_uuid(dashed=False))
        # Add 6 rows to table
        for uuidstr in self.uuidstrs:
            ins_stmt = self.plans.insert().values(id=uuidstr)
            self.conn.execute(ins_stmt)
            ins_stmt = self.resources.insert().values(plan_id=uuidstr)
            self.conn.execute(ins_stmt)

        # Set 4 of them deleted, 2 are 60 days ago, 2 are 20 days ago
        old = timeutils.utcnow() - datetime.timedelta(days=20)
        older = timeutils.utcnow() - datetime.timedelta(days=60)
        make_plans_old = self.plans.update().where(
            self.plans.c.id.in_(self.uuidstrs[1:3])).values(
            deleted_at=old)
        make_plans_older = self.plans.update().where(
            self.plans.c.id.in_(self.uuidstrs[4:6])).values(
            deleted_at=older)
        make_resources_old = self.resources.update().where(
            self.resources.c.plan_id.in_(self.uuidstrs[1:3])).values(
            deleted_at=old)
        make_resources_older = self.resources.update().where(
            self.resources.c.plan_id.in_(self.uuidstrs[4:6])).values(
            deleted_at=older)

        self.conn.execute(make_plans_old)
        self.conn.execute(make_plans_older)
        self.conn.execute(make_resources_old)
        self.conn.execute(make_resources_older)

    def test_purge_deleted_rows_old(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            # We're seeing issues with foreign key support in SQLite 3.6.20
            # SQLAlchemy doesn't support it at all with < SQLite 3.6.19
            # It works fine in SQLite 3.7.
            # Force foreign_key checking if running SQLite >= 3.7
            import sqlite3
            tup = sqlite3.sqlite_version_info
            if tup[0] > 3 or (tup[0] == 3 and tup[1] >= 7):
                self.conn.execute("PRAGMA foreign_keys = ON")
        # Purge at 30 days old, should only delete 2 rows
        db.purge_deleted_rows(self.context, age_in_days=30)
        plans_rows = self.session.query(self.plans).count()
        resources_rows = self.session.query(self.resources).count()
        # Verify that we only deleted 2
        self.assertEqual(4, plans_rows)
        self.assertEqual(4, resources_rows)

    def test_purge_deleted_rows_older(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            # We're seeing issues with foreign key support in SQLite 3.6.20
            # SQLAlchemy doesn't support it at all with < SQLite 3.6.19
            # It works fine in SQLite 3.7.
            # Force foreign_key checking if running SQLite >= 3.7
            import sqlite3
            tup = sqlite3.sqlite_version_info
            if tup[0] > 3 or (tup[0] == 3 and tup[1] >= 7):
                self.conn.execute("PRAGMA foreign_keys = ON")
        # Purge at 10 days old now, should delete 2 more rows
        db.purge_deleted_rows(self.context, age_in_days=10)

        plans_rows = self.session.query(self.plans).count()
        resources_rows = self.session.query(self.resources).count()
        # Verify that we only have 2 rows now
        self.assertEqual(2, plans_rows)
        self.assertEqual(2, resources_rows)

    def test_purge_deleted_rows_bad_args(self):
        # Test with no age argument
        self.assertRaises(TypeError, db.purge_deleted_rows, self.context)
        # Test purge with non-integer
        self.assertRaises(exception.InvalidParameterValue,
                          db.purge_deleted_rows, self.context,
                          age_in_days='ten')

    def test_purge_deleted_rows_integrity_failure(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            # We're seeing issues with foreign key support in SQLite 3.6.20
            # SQLAlchemy doesn't support it at all with < SQLite 3.6.19
            # It works fine in SQLite 3.7.
            # So return early to skip this test if running SQLite < 3.7
            import sqlite3
            tup = sqlite3.sqlite_version_info
            if tup[0] < 3 or (tup[0] == 3 and tup[1] < 7):
                self.skipTest(
                    'sqlite version too old for reliable SQLA foreign_keys')
            self.conn.execute("PRAGMA foreign_keys = ON")

        # add new entry in plans and resources for
        # integrity check
        uuid_str = uuidutils.generate_uuid(dashed=False)
        ins_stmt = self.plans.insert().values(id=uuid_str)
        self.conn.execute(ins_stmt)
        ins_stmt = self.resources.insert().values(
            plan_id=uuid_str)
        self.conn.execute(ins_stmt)

        # set plans record to deleted 20 days ago
        old = timeutils.utcnow() - datetime.timedelta(days=20)
        make_old = self.plans.update().where(
            self.plans.c.id.in_([uuid_str])).values(deleted_at=old)
        self.conn.execute(make_old)

        # Verify that purge_deleted_rows fails due to Foreign Key constraint
        self.assertRaises(db_exc.DBReferenceError, db.purge_deleted_rows,
                          self.context, age_in_days=10)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
SQLAlchemy models for karbor data.
"""

from oslo_config import cfg
from oslo_db.sqlalchemy import models
from oslo_utils import timeutils
from sqlalchemy import Column, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import DateTime, Boolean, ForeignKey
from sqlalchemy import orm

CONF = cfg.CONF
BASE = declarative_base()


class KarborBase(models.TimestampMixin,
                 models.ModelBase):
    """Base class for karbor Models."""

    __table_args__ = {'mysql_engine': 'InnoDB'}

    deleted_at = Column(DateTime)
    deleted = Column(Boolean, default=False)
    metadata = None

    def delete(self, session):
        """Delete this object."""
        self.deleted = True
        self.deleted_at = timeutils.utcnow()
        self.save(session=session)


class Service(BASE, KarborBase):
    """Represents a running service on a host."""

    __tablename__ = 'services'
    id = Column(Integer, primary_key=True)
    host = Column(String(255))  # , ForeignKey('hosts.id'))
    binary = Column(String(255))
    topic = Column(String(255))
    report_count = Column(Integer, nullable=False, default=0)
    disabled = Column(Boolean, default=False)
    disabled_reason = Column(String(255))
    # adding column modified_at to contain timestamp
    # for manual enable/disable of karbor services
    # updated_at column will now contain timestamps for
    # periodic updates
    modified_at = Column(DateTime)
    rpc_current_version = Column(String(36))
    rpc_available_version = Column(String(36))


class Trigger(BASE, KarborBase):
    """Represents a trigger."""

    __tablename__ = 'triggers'

    id = Column(String(36), primary_key=True, nullable=False)
    name = Column(String(255), nullable=False)
    project_id = Column(String(255), nullable=False)
    type = Column(String(64), nullable=False)
    properties = Column(Text, nullable=False)


class ScheduledOperation(BASE, KarborBase):
    """Represents a scheduled operation."""

    __tablename__ = 'scheduled_operations'

    id = Column(String(36), primary_key=True, nullable=False)
    name = Column(String(255), nullable=False)
    description = Column(String(255))
    operation_type = Column(String(64), nullable=False)
    user_id = Column(String(64), nullable=False)
    project_id = Column(String(255), nullable=False)
    trigger_id = Column(String(36), ForeignKey('triggers.id'),
                        index=True, nullable=False)
    operation_definition = Column(Text, nullable=False)
    enabled = Column(Boolean, default=True)

    trigger = orm.relationship(
        Trigger,
        foreign_keys=trigger_id,
        primaryjoin='and_('
                    'ScheduledOperation.trigger_id == Trigger.id,'
                    'ScheduledOperation.deleted == 0,'
                    'Trigger.deleted == 0)')


class ScheduledOperationState(BASE, KarborBase):
    """Represents a scheduled operation state."""

    __tablename__ = 'scheduled_operation_states'

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    operation_id = Column(String(36),
                          ForeignKey('scheduled_operations.id',
                                     ondelete='CASCADE'),
                          index=True, unique=True,
                          nullable=False)
    service_id = Column(Integer, ForeignKey('services.id'), nullable=False)
    trust_id = Column(String(64), nullable=False)
    state = Column(String(32), nullable=False)
    end_time_for_run = Column(DateTime)

    operation = orm.relationship(
        ScheduledOperation,
        foreign_keys=operation_id,
        primaryjoin='and_('
                    'ScheduledOperationState.operation_id == '
                    'ScheduledOperation.id,'
                    'ScheduledOperationState.deleted == 0,'
                    'ScheduledOperation.deleted == 0)')


class ScheduledOperationLog(BASE, KarborBase):
    """Represents a scheduled operation log."""

    __tablename__ = 'scheduled_operation_logs'

    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    operation_id = Column(String(36),
                          ForeignKey('scheduled_operations.id',
                                     ondelete='CASCADE'),
                          index=True, nullable=False)
    expect_start_time = Column(DateTime)
    triggered_time = Column(DateTime)
    actual_start_time = Column(DateTime)
    end_time = Column(DateTime)
    state = Column(String(32), nullable=False)
    extend_info = Column(Text)


class Plan(BASE, KarborBase):
    """Represents a Plan."""

    __tablename__ = 'plans'
    id = Column(String(36), primary_key=True)
    name = Column(String(255))
    description = Column(String(255))
    provider_id = Column(String(36))
    project_id = Column(String(255))
    status = Column(String(64))
    parameters = Column(Text)


class Resource(BASE, KarborBase):
    """Represents a resource in a plan."""

    __tablename__ = 'resources'
    id = Column(Integer, primary_key=True)
    resource_id = Column(String(36))
    resource_type = Column(String(64))
    resource_name = Column(String(255))
    resource_extra_info = Column(Text)
    plan_id = Column(String(36), ForeignKey('plans.id'), nullable=False)
    plan = orm.relationship(Plan, backref="resources",
                            foreign_keys=plan_id,
                            primaryjoin='and_('
                            'Resource.plan_id == Plan.id,'
                            'Resource.deleted == False)')


class Restore(BASE, KarborBase):
    """Represents a Restore."""

    __tablename__ = 'restores'
    id = Column(String(36), primary_key=True)
    project_id = Column(String(255))
    provider_id = Column(String(36))
    checkpoint_id = Column(String(36))
    restore_target = Column(String(255))
    parameters = Column(Text)
    status = Column(String(64))
    resources_status = Column(Text)
    resources_reason = Column(Text)


class OperationLog(BASE, KarborBase):
    """Represents a operation log."""

    __tablename__ = 'operation_logs'
    id = Column(String(36), primary_key=True)
    project_id = Column(String(255))
    scheduled_operation_id = Column(String(36))
    started_at = Column(DateTime)
    ended_at = Column(DateTime)
    state = Column(String(64))
    error = Column(String(64))
    entries = Column(Text)


class CheckpointRecord(BASE, KarborBase):
    """Represents a checkpoint record."""

    __tablename__ = 'checkpoint_records'

    id = Column(String(36), primary_key=True, nullable=False)
    project_id = Column(String(36), nullable=False)
    checkpoint_id = Column(String(36), nullable=False)
    checkpoint_status = Column(String(36), nullable=False)
    provider_id = Column(String(36), nullable=False)
    plan_id = Column(String(36), nullable=False)
    operation_id = Column(String(36))
    create_by = Column(String(36))
    extend_info = Column(Text)


def register_models():
    """Register Models and create metadata.

    Called from karbor.db.sqlalchemy.__init__ as part of loading the driver,
    it will never need to be called explicitly elsewhere unless the
    connection is lost and needs to be reestablished.
    """
    from sqlalchemy import create_engine
    models = (Service,
              Plan,
              Resource,
              Trigger,
              ScheduledOperation,
              ScheduledOperationState,
              ScheduledOperationLog,
              Restore,
              CheckpointRecord)
    engine = create_engine(CONF.database.connection, echo=False)
    for model in models:
        model.metadata.create_all(engine)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from karbor.tests.fullstack import karbor_base


class ProvidersTest(karbor_base.KarborBaseTest):
    """Test Providers operation"""

    def test_providers_list(self):
        provider_res = self.karbor_client.providers.list()
        self.assertTrue(len(provider_res))

    def test_provider_get(self):
        providers = self.karbor_client.providers.list()
        for provider in providers:
            provider_res = self.karbor_client.providers.get(provider.id)
            self.assertEqual(provider.name, provider_res.name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from karbor.common import karbor_keystone_plugin


LOG = logging.getLogger(__name__)


class UserTrustManager(object):
    def __init__(self):
        super(UserTrustManager, self).__init__()
        self._user_trust_map = {}
        self._skp = karbor_keystone_plugin.KarborKeystonePlugin()

    def _user_trust_key(self, user_id, project_id):
        return "%s_%s" % (user_id, project_id)

    def _add_user_trust_info(self, user_id, project_id,
                             operation_id, trust_id, session):
        key = self._user_trust_key(user_id, project_id)
        self._user_trust_map[key] = {
            'operation_ids': {operation_id},
            'trust_id': trust_id,
            'session': session
        }

    def _get_user_trust_info(self, user_id, project_id):
        return self._user_trust_map.get(
            self._user_trust_key(user_id, project_id))

    def _del_user_trust_info(self, user_id, project_id):
        key = self._user_trust_key(user_id, project_id)
        del self._user_trust_map[key]

    def get_token(self, user_id, project_id):
        auth_info = self._get_user_trust_info(user_id, project_id)
        if not auth_info:
            return None

        try:
            return auth_info['session'].get_token()
        except Exception:
            LOG.exception("Get token failed, user_id=%(user_id)s, "
                          "project_id=%(proj_id)s",
                          {'user_id': user_id, 'proj_id': project_id})
        return None

    def add_operation(self, context, operation_id):
        auth_info = self._get_user_trust_info(
            context.user_id, context.project_id)
        if auth_info:
            auth_info['operation_ids'].add(operation_id)
            return auth_info['trust_id']

        trust_id = self._skp.create_trust_to_karbor(context)
        try:
            lsession = self._skp.create_trust_session(trust_id)
        except Exception:
            self._skp.delete_trust_to_karbor(trust_id)
            raise

        self._add_user_trust_info(context.user_id, context.project_id,
                                  operation_id, trust_id, lsession)

        return trust_id

    def delete_operation(self, context, operation_id):
        auth_info = self._get_user_trust_info(
            context.user_id, context.project_id)
        if not auth_info:
            return

        operation_ids = auth_info['operation_ids']
        operation_ids.discard(operation_id)
        if len(operation_ids) == 0:
            self._skp.delete_trust_to_karbor(auth_info['trust_id'])
            self._del_user_trust_info(context.user_id, context.project_id)

    def resume_operation(self, operation_id, user_id, project_id, trust_id):
        auth_info = self._get_user_trust_info(user_id, project_id)
        if auth_info:
            auth_info['operation_ids'].add(operation_id)
            return

        try:
            lsession = self._skp.create_trust_session(trust_id)
        except Exception:
            raise

        self._add_user_trust_info(user_id, project_id,
                                  operation_id, trust_id, lsession)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from karbor.common import constants
from karbor import resource
from karbor.services.protection import protectable_plugin


class ProjectProtectablePlugin(protectable_plugin.ProtectablePlugin):
    """Keystone project protectable plugin"""
    _SUPPORT_RESOURCE_TYPE = constants.PROJECT_RESOURCE_TYPE

    def get_resource_type(self):
        return self._SUPPORT_RESOURCE_TYPE

    def get_parent_resource_types(self):
        return ()

    def list_resources(self, context, parameters=None):
        # TODO(yuvalbr) handle admin context for multiple projects?
        return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                  id=context.project_id,
                                  name=context.project_name)]

    def get_dependent_resources(self, context, parent_resource):
        pass

    def show_resource(self, context, resource_id, parameters=None):
        # TODO(yinwei) get project name through keystone client
        return resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                 id=resource_id,
                                 name=context.project_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial

from novaclient import exceptions
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import constants
from karbor import exception
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protection_plugin
from karbor.services.protection.protection_plugins.server \
    import server_plugin_schemas
from karbor.services.protection.protection_plugins import utils


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

VOLUME_ATTACHMENT_RESOURCE = 'OS::Cinder::VolumeAttachment'
FLOATING_IP_ASSOCIATION = 'OS::Nova::FloatingIPAssociation'

nova_backup_opts = [
    cfg.IntOpt(
        'poll_interval', default=15,
        help='Poll interval for Nova backup status'
    ),
]


class ProtectOperation(protection_plugin.Operation):
    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        server_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(server_id)

        nova_client = ClientFactory.create_client("nova", context)
        cinder_client = ClientFactory.create_client("cinder", context)
        neutron_client = ClientFactory.create_client("neutron", context)

        resource_definition = {"resource_id": server_id}

        # get dependent resources
        server_child_nodes = []
        resources = checkpoint.resource_graph
        for resource_node in resources:
            resource = resource_node.value
            if resource.id == server_id:
                server_child_nodes = resource_node.child_nodes

        LOG.info("Creating server backup, server_id: %s. ", server_id)
        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_PROTECTING)

            # get attach_metadata about volume
            attach_metadata = {}
            for server_child_node in server_child_nodes:
                child_resource = server_child_node.value
                if child_resource.type == constants.VOLUME_RESOURCE_TYPE:
                    volume = cinder_client.volumes.get(child_resource.id)
                    attachments = getattr(volume, "attachments")
                    for attachment in attachments:
                        if attachment["server_id"] == server_id:
                            attachment["bootable"] = getattr(
                                volume, "bootable")
                            attach_metadata[child_resource.id] = attachment
            resource_definition["attach_metadata"] = attach_metadata

            # get metadata about AZ
            server = nova_client.servers.get(server_id)
            availability_zone = getattr(server, "OS-EXT-AZ:availability_zone")

            # get metadata about network, flavor, key_name, security_groups
            addresses = getattr(server, "addresses")
            networks = []
            floating_ips = []
            for network_infos in addresses.values():
                for network_info in network_infos:
                    addr = network_info.get("addr")
                    mac = network_info.get("OS-EXT-IPS-MAC:mac_addr")
                    network_type = network_info.get("OS-EXT-IPS:type")
                    if network_type == 'fixed':
                        port = neutron_client.list_ports(
                            mac_address=mac)["ports"][0]
                        if port["network_id"] not in networks:
                            networks.append(port["network_id"])
                    elif network_type == "floating":
                        floating_ips.append(addr)
            flavor = getattr(server, "flavor")["id"]
            key_name = getattr(server, "key_name", None)
            security_groups = getattr(server, "security_groups", None)

            # get metadata about boot device
            boot_metadata = {}
            image_info = getattr(server, "image", None)
            if image_info is not None and isinstance(image_info, dict):
                boot_metadata["boot_device_type"] = "image"
                boot_metadata["boot_image_id"] = image_info['id']
            else:
                boot_metadata["boot_device_type"] = "volume"
                volumes_attached = getattr(
                    server, "os-extended-volumes:volumes_attached", [])
                for volume_attached in volumes_attached:
                    volume_id = volume_attached["id"]
                    volume_attach_metadata = attach_metadata.get(
                        volume_id, None)
                    if volume_attach_metadata is not None and (
                            volume_attach_metadata["bootable"] == "true"):
                        boot_metadata["boot_volume_id"] = volume_id
                        boot_metadata["boot_attach_metadata"] = (
                            volume_attach_metadata)
            resource_definition["boot_metadata"] = boot_metadata

            # save all server's metadata
            server_metadata = {"availability_zone": availability_zone,
                               "networks": networks,
                               "floating_ips": floating_ips,
                               "flavor": flavor,
                               "key_name": key_name,
                               "security_groups": security_groups,
                               }
            resource_definition["server_metadata"] = server_metadata
            LOG.info("Creating server backup, resource_definition: %s.",
                     resource_definition)
            bank_section.update_object("metadata", resource_definition)

            # update resource_definition backup_status
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_AVAILABLE)
            LOG.info("Finish backup server, server_id: %s.", server_id)
        except Exception as err:
            # update resource_definition backup_status
            LOG.exception("Create backup failed, server_id: %s.", server_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason=err,
                resource_id=server_id,
                resource_type=constants.SERVER_RESOURCE_TYPE)


class DeleteOperation(protection_plugin.Operation):
    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        resource_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(resource_id)

        LOG.info("deleting server backup, server_id: %s.", resource_id)

        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETING)
            objects = bank_section.list_objects()
            for obj in objects:
                if obj == "status":
                    continue
                bank_section.delete_object(obj)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETED)
            LOG.info("finish delete server, server_id: %s.", resource_id)
        except Exception as err:
            # update resource_definition backup_status
            LOG.error("Delete backup failed, server_id: %s.", resource_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.DeleteBackupFailed(
                reason=err,
                resource_id=resource_id,
                resource_type=constants.SERVER_RESOURCE_TYPE)


class RestoreOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(RestoreOperation, self).__init__()
        self._interval = poll_interval

    def on_complete(self, checkpoint, resource, context, parameters, **kwargs):
        original_server_id = resource.id
        LOG.info("Restoring server backup, server_id: %s.", original_server_id)

        update_method = None
        try:
            resource_definition = checkpoint.get_resource_bank_section(
                original_server_id).get_object("metadata")

            nova_client = ClientFactory.create_client("nova", context)
            heat_template = kwargs.get("heat_template")

            # restore server instance
            new_server_id = self._restore_server_instance(
                nova_client, heat_template, original_server_id,
                parameters.get("restore_name", "karbor-restore-server"),
                resource_definition)

            update_method = partial(utils.udpate_resource_restore_result,
                                    kwargs.get('restore'), resource.type,
                                    new_server_id)
            update_method(constants.RESOURCE_STATUS_RESTORING)
            self._wait_server_to_active(nova_client, new_server_id)

            # restore volume attachment
            self._restore_volume_attachment(
                nova_client, ClientFactory.create_client("cinder", context),
                heat_template, new_server_id, resource_definition)

            # restore floating ip association
            self._restore_floating_association(
                nova_client, new_server_id, resource_definition)

            heat_template.put_parameter(original_server_id, new_server_id)

            update_method(constants.RESOURCE_STATUS_AVAILABLE)

            LOG.info("Finish restore server, server_id: %s.",
                     original_server_id)

        except Exception as e:
            if update_method:
                update_method(constants.RESOURCE_STATUS_ERROR, str(e))
            LOG.exception("Restore server backup failed, server_id: %s.",
                          original_server_id)
            raise exception.RestoreBackupFailed(
                reason=e,
                resource_id=original_server_id,
                resource_type=constants.SERVER_RESOURCE_TYPE
            )

    def _restore_server_instance(self, nova_client, heat_template,
                                 original_id, restore_name,
                                 resource_definition):
        server_metadata = resource_definition["server_metadata"]
        properties = {
            "availability_zone": server_metadata.get("availability_zone"),
            "flavor": server_metadata.get("flavor"),
            "name": restore_name,
            "image": None
        }

        # server boot device
        boot_metadata = resource_definition["boot_metadata"]
        boot_device_type = boot_metadata.get("boot_device_type")
        if boot_device_type == "image":
            properties["image"] = heat_template.get_resource_reference(
                boot_metadata.get("boot_image_id"))

        elif boot_device_type == "volume":
            properties["block_device_mapping_v2"] = [{
                'uuid': heat_template.get_resource_reference(
                    boot_metadata.get("boot_volume_id")),
                'source_type': 'volume',
                'destination_type': 'volume',
                'boot_index': 0,
                'delete_on_termination': False,
            }]
        else:
            reason = "Can not find the boot device of the server."
            LOG.error("Restore server backup failed, (server_id:"
                      "%(server_id)s): %(reason)s.",
                      {'server_id': original_id,
                       'reason': reason})
            raise Exception(reason)

        # server key_name, security_groups, networks
        properties["key_name"] = server_metadata.get("key_name", None)

        if server_metadata.get("security_groups"):
            properties["security_groups"] = [
                security_group["name"]
                for security_group in server_metadata["security_groups"]
            ]

        if server_metadata.get("networks"):
            properties["nics"] = [
                {'net-id': network}
                for network in server_metadata["networks"]
            ]

        properties["userdata"] = None

        try:
            server = nova_client.servers.create(**properties)
        except Exception as ex:
            LOG.error('Error creating server (server_id:%(server_id)s): '
                      '%(reason)s',
                      {'server_id': original_id,
                       'reason': ex})
            raise

        return server.id

    def _restore_volume_attachment(self, nova_client, cinder_client,
                                   heat_template, new_server_id,
                                   resource_definition):
        attach_metadata = resource_definition.get("attach_metadata", {})
        for original_id, attach_metadata_item in attach_metadata.items():
            if attach_metadata_item.get("bootable", None) == "true":
                continue

            volume_id = heat_template.get_resource_reference(original_id)
            try:
                nova_client.volumes.create_server_volume(
                    server_id=new_server_id,
                    volume_id=volume_id,
                    device=attach_metadata_item.get("device", None))

            except Exception as ex:
                LOG.error("Failed to attach volume %(vol)s to server %(srv)s, "
                          "reason: %(err)s",
                          {'vol': volume_id,
                           'srv': new_server_id,
                           'err': ex})
                raise

            self._wait_volume_to_attached(cinder_client, volume_id)

    def _restore_floating_association(self, nova_client, new_server_id,
                                      resource_definition):
        server_metadata = resource_definition["server_metadata"]
        for floating_ip in server_metadata.get("floating_ips", []):
            nova_client.servers.add_floating_ip(
                nova_client.servers.get(new_server_id), floating_ip)

    def _wait_volume_to_attached(self, cinder_client, volume_id):
        def _get_volume_status():
            try:
                return cinder_client.volumes.get(volume_id).status
            except Exception as ex:
                LOG.error('Fetch volume(%(volume_id)s) status failed, '
                          'reason: %(reason)s',
                          {'volume_id': volume_id,
                           'reason': ex})
                return 'ERROR'

        is_success = utils.status_poll(
            _get_volume_status,
            interval=self._interval,
            success_statuses={'in-use', },
            failure_statuses={'ERROR', },
            ignore_statuses={'available', 'attaching'}
        )
        if not is_success:
            raise Exception('Attach the volume to server failed')

    def _wait_server_to_active(self, nova_client, server_id):
        def _get_server_status():
            try:
                server = self._fetch_server(nova_client, server_id)
                return server.status.split('(')[0] if server else 'BUILD'
            except Exception as ex:
                LOG.error('Fetch server(%(server_id)s) failed, '
                          'reason: %(reason)s',
                          {'server_id': server_id,
                           'reason': ex})
                return 'ERROR'

        is_success = utils.status_poll(
            _get_server_status,
            interval=self._interval,
            success_statuses={'ACTIVE', },
            failure_statuses={'ERROR', },
            ignore_statuses={'BUILD', 'HARD_REBOOT', 'PASSWORD', 'REBOOT',
                             'RESCUE', 'RESIZE', 'REVERT_RESIZE', 'SHUTOFF',
                             'SUSPENDED', 'VERIFY_RESIZE'},
        )
        if not is_success:
            raise Exception('The server does not start successfully')

    def _fetch_server(self, nova_client, server_id):
        server = None
        try:
            server = nova_client.servers.get(server_id)
        except exceptions.OverLimit as exc:
            LOG.warning("Received an OverLimit response when "
                        "fetching server (%(id)s) : %(exception)s",
                        {'id': server_id,
                         'exception': exc})
        except exceptions.ClientException as exc:
            if ((getattr(exc, 'http_status', getattr(exc, 'code', None)) in
                 (500, 503))):
                LOG.warning("Received the following exception when "
                            "fetching server (%(id)s) : %(exception)s",
                            {'id': server_id,
                             'exception': exc})
            else:
                raise
        return server


class NovaProtectionPlugin(protection_plugin.ProtectionPlugin):
    _SUPPORT_RESOURCE_TYPES = [constants.SERVER_RESOURCE_TYPE]

    def __init__(self, config=None):
        super(NovaProtectionPlugin, self).__init__(config)
        self._config.register_opts(nova_backup_opts,
                                   'nova_backup_protection_plugin')
        self._poll_interval = (
            self._config.nova_backup_protection_plugin.poll_interval)

    @classmethod
    def get_supported_resources_types(cls):
        return cls._SUPPORT_RESOURCE_TYPES

    @classmethod
    def get_options_schema(cls, resource_type):
        return server_plugin_schemas.OPTIONS_SCHEMA

    @classmethod
    def get_restore_schema(cls, resource_type):
        return server_plugin_schemas.RESTORE_SCHEMA

    @classmethod
    def get_saved_info_schema(cls, resource_type):
        return server_plugin_schemas.SAVED_INFO_SCHEMA

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        pass

    def get_protect_operation(self, resource):
        return ProtectOperation()

    def get_restore_operation(self, resource):
        return RestoreOperation(self._poll_interval)

    def get_delete_operation(self, resource):
        return DeleteOperation()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_versionedobjects import fields

from karbor import db
from karbor.objects import base


@base.KarborObjectRegistry.register
class Trigger(base.KarborPersistentObject, base.KarborObject,
              base.KarborObjectDictCompat,
              base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(),
        'project_id': fields.StringField(),
        'type': fields.StringField(),
        'properties': fields.DictOfStringsField(),
    }

    @staticmethod
    def _from_db_object(context, trigger, db_trigger):
        special_fields = set(['properties'])

        normal_fields = set(trigger.fields) - special_fields
        for name in normal_fields:
            trigger[name] = db_trigger.get(name)

        properties = db_trigger['properties']
        if properties:
            trigger['properties'] = jsonutils.loads(properties)

        trigger._context = context
        trigger.obj_reset_changes()
        return trigger

    @staticmethod
    def _convert_properties_to_db_format(updates):
        properties = updates.pop('properties', None)
        if properties is not None:
            updates['properties'] = jsonutils.dumps(properties)

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_trigger = db.trigger_get(context, id)
        if db_trigger:
            return cls._from_db_object(context, cls(), db_trigger)

    @base.remotable
    def create(self):
        updates = self.karbor_obj_get_changes()
        self._convert_properties_to_db_format(updates)
        db_trigger = db.trigger_create(self._context, updates)
        self._from_db_object(self._context, self, db_trigger)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates and self.id:
            self._convert_properties_to_db_format(updates)
            db.trigger_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        if self.id:
            db.trigger_delete(self._context, self.id)


@base.KarborObjectRegistry.register
class TriggerList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('Trigger'),
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, filters, limit=None,
                       marker=None, sort_keys=None, sort_dirs=None):

        db_trigger_list = db.trigger_get_all_by_filters_sort(
            context, filters, limit=limit, marker=marker,
            sort_keys=sort_keys, sort_dirs=sort_dirs)

        return base.obj_make_list(context, cls(context), Trigger,
                                  db_trigger_list)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from datetime import datetime
from oslo_serialization import jsonutils

from karbor import exception
from karbor.services.operationengine.engine.triggers.timetrigger.timeformats \
    import calendar_time
from karbor.tests import base


class CalendarTimeTestCase(base.TestCase):

    def test_invalid_pattern(self):
        patterns = [
            "     ",

            "DTSTART:20070220T170000Z\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=1\n",

            "BEGIN:VCALENDAR\n"
            "BEGIN:VEVENT\n"
            "DTSTART:20070220T170000\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=6\n"
            "END:VCALENDAR",

            "BEGIN:VCALENDAR\n"
            "DTSTART:20070220T170000\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=6\n"
            "END:VEVENT\n"
            "END:VCALENDAR",
        ]

        regexp = re.compile("^The trigger pattern.* is invalid$", re.DOTALL)
        for pattern in patterns:
            self.assertRaisesRegex(exception.InvalidInput,
                                   regexp,
                                   calendar_time.ICal.check_time_format,
                                   pattern)

        patterns = [
            "BEGIN:VCALENDAR\n"
            "END:VCALENDAR",

            "BEGIN:VCALENDAR\n"
            "DTSTART:20070220T170000\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=6\n"
            "END:VCALENDAR",

            "BEGIN:VCALENDAR\n"
            "BEGIN:VTODO\n"
            "END:VTODO\n"
            "END:VCALENDAR",
        ]

        regexp = re.compile("^The trigger pattern.* must include less than "
                            "one VEVENT component$", re.DOTALL)
        for pattern in patterns:
            self.assertRaisesRegex(exception.InvalidInput,
                                   regexp,
                                   calendar_time.ICal.check_time_format,
                                   pattern)

        patterns = [
            "BEGIN:VEVENT\n"
            "END:VEVENT",

            "BEGIN:VCALENDAR\n"
            "BEGIN:VEVENT\n"
            "END:VEVENT\n"
            "END:VCALENDAR",

            "BEGIN:VEVENT\n"
            "DTSTART:20070220T170000Z\n"
            "END:VEVENT",
        ]

        regexp = re.compile("^The first VEVENT component of trigger pattern.* "
                            "must include less than one RRULE property$",
                            re.DOTALL)
        for pattern in patterns:
            self.assertRaisesRegex(exception.InvalidInput,
                                   regexp,
                                   calendar_time.ICal.check_time_format,
                                   pattern)

    def test_valid_pattern(self):
        pattern = "BEGIN:VEVENT\nRRULE:FREQ=MINUTELY;INTERVAL=60;\nEND:VEVENT"
        self.assertIsNone(calendar_time.ICal.check_time_format(pattern))

    def test_escape_valid_pattern(self):
        pattern0 = "BEGIN:VEVENT\\nRRULE:FREQ=HOURLY;INTERVAL=1;\\nEND:VEVENT"
        self.assertIsNone(calendar_time.ICal.check_time_format(pattern0))

        pattern1 = "BEGIN:VEVENT\nRRULE:FREQ=HOURLY;INTERVAL=1;\nEND:VEVENT"
        properties = {"format": "calendar",
                      "pattern": pattern1}
        body = {"trigger_info": {"name": "test",
                                 "type": "time",
                                 "properties": properties,
                                 }}
        quest = jsonutils.dumps(body)
        recieve = jsonutils.loads(quest)
        trigger_info = recieve["trigger_info"]
        trigger_property = trigger_info.get("properties", None)
        pattern_ = trigger_property.get("pattern", None)

        self.assertIsNone(calendar_time.ICal.check_time_format(pattern_))

    def test_compute_next_time(self):
        pattern = (
            "BEGIN:VEVENT\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=1\n"
            "END:VEVENT"
        )
        dtstart = datetime(2016, 2, 20, 17, 0, 0)
        time_obj = calendar_time.ICal(dtstart, pattern)
        now = datetime(2016, 2, 20, 15, 11, 0)
        time1 = time_obj.compute_next_time(now)
        time2 = datetime(2016, 2, 20, 17, 1, 0)
        self.assertEqual(time2, time1)
        now = datetime(2016, 3, 20, 15, 11, 0)
        time1 = time_obj.compute_next_time(now)
        time2 = datetime(2016, 3, 26, 17, 1, 0)
        self.assertEqual(time2, time1)

        pattern = (
            "BEGIN:VEVENT\n"
            "RRULE:FREQ=WEEKLY;BYDAY=MO,WE,FR;BYHOUR=10;BYMINUTE=0\n"
            "RRULE:FREQ=WEEKLY;BYDAY=TU,TH,SA;BYHOUR=20;BYMINUTE=0\n"
            "END:VEVENT"
        )
        dtstart = datetime(2016, 2, 20, 17, 0, 0)
        time_obj = calendar_time.ICal(dtstart, pattern)
        now = datetime(2016, 7, 31, 15, 11, 0)
        time1 = time_obj.compute_next_time(now)
        time2 = datetime(2016, 8, 1, 10, 0, 0)
        self.assertEqual(time2, time1)
        time1 = time_obj.compute_next_time(time2)
        time2 = datetime(2016, 8, 2, 20, 0, 0)
        self.assertEqual(time2, time1)
        time1 = time_obj.compute_next_time(time2)
        time2 = datetime(2016, 8, 3, 10, 0, 0)
        self.assertEqual(time2, time1)

    def test_get_min_interval(self):
        pattern = (
            "BEGIN:VEVENT\n"
            "RRULE:FREQ=WEEKLY;INTERVAL=1;BYHOUR=17;BYMINUTE=1\n"
            "END:VEVENT"
        )
        dtstart = datetime(2016, 2, 20, 17, 0, 0)
        time_obj = calendar_time.ICal(dtstart, pattern)
        self.assertEqual(604800, time_obj.get_min_interval())

        pattern = (
            "BEGIN:VEVENT\n"
            "RRULE:FREQ=WEEKLY;COUNT=1\n"
            "END:VEVENT"
        )
        dtstart = datetime(2016, 2, 20, 17, 0, 0)
        time_obj = calendar_time.ICal(dtstart, pattern)
        self.assertIsNone(time_obj.get_min_interval())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from karbor import objects
from karbor.tests.unit import objects as test_objects

NOW = timeutils.utcnow().replace(microsecond=0)

Operation_ID = '0354ca9ddcd046b693340d78759fd274'

Fake_Operation = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': NOW,
    'deleted': False,
    'id': Operation_ID,
    'name': 'protect vm',
    'description': 'protect vm resource',
    'operation_type': 'protect',
    'user_id': '123',
    'project_id': '123',
    'trigger_id': '0354ca9ddcd046b693340d78759fd275',
    'operation_definition': '{}'
}


class TestScheduledOperation(test_objects.BaseObjectsTestCase):
    Operation_Class = objects.ScheduledOperation

    @mock.patch('karbor.db.scheduled_operation_get')
    def test_get_by_id(self, operation_get):
        db_op = Fake_Operation.copy()
        operation_get.return_value = db_op

        op = self.Operation_Class.get_by_id(self.context, Operation_ID)
        db_op['operation_definition'] = jsonutils.loads(
            db_op['operation_definition'])
        self._compare(self, db_op, op)
        operation_get.assert_called_once_with(self.context, Operation_ID, [])

    @mock.patch('karbor.db.scheduled_operation_get')
    def test_get_join_trigger(self, operation_get):
        db_op = Fake_Operation.copy()
        db_op['trigger'] = {
            'created_at': NOW,
            'deleted_at': None,
            'updated_at': NOW,
            'deleted': False,
            'id': '123',
            'name': 'daily',
            'project_id': '123',
            'type': 'time',
            'properties': '{}'
        }
        operation_get.return_value = db_op

        op = self.Operation_Class.get_by_id(self.context,
                                            Operation_ID, ['trigger'])
        db_op['operation_definition'] = jsonutils.loads(
            db_op['operation_definition'])
        self.assertEqual(db_op['trigger']['type'], op.trigger.type)
        operation_get.assert_called_once_with(self.context,
                                              Operation_ID, ['trigger'])

    @mock.patch('karbor.db.scheduled_operation_create')
    def test_create(self, operation_create):
        db_op = Fake_Operation.copy()
        operation_create.return_value = db_op

        op = self.Operation_Class(context=self.context)
        op.create()
        db_op['operation_definition'] = jsonutils.loads(
            db_op['operation_definition'])
        self._compare(self, db_op, op)
        operation_create.assert_called_once_with(self.context, {})

    @mock.patch('karbor.db.scheduled_operation_update')
    def test_save(self, operation_update):
        db_op = Fake_Operation
        op = self.Operation_Class._from_db_object(self.context,
                                                  self.Operation_Class(),
                                                  db_op)
        fake_op_def = {'a': '1'}
        op.name = 'protect volume'
        op.operation_definition = fake_op_def
        op.save()

        operation_update.assert_called_once_with(
            self.context, op.id,
            {'name': 'protect volume',
             'operation_definition': jsonutils.dumps(fake_op_def)})

    @mock.patch('karbor.db.scheduled_operation_delete')
    def test_destroy(self, operation_delete):
        db_op = Fake_Operation
        op = self.Operation_Class._from_db_object(self.context,
                                                  self.Operation_Class(),
                                                  db_op)
        op.destroy()
        operation_delete.assert_called_once_with(self.context, op.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import yaml

from karbor.exception import InvalidOriginalId
from oslo_log import log as logging
from oslo_serialization import jsonutils

LOG = logging.getLogger(__name__)


class HeatResource(object):
    def __init__(self, resource_id, type):
        super(HeatResource, self).__init__()
        self.resource_id = resource_id
        self._type = type
        self._properties = {}

    def set_property(self, key, value):
        self._properties[key] = value

    def to_dict(self):
        resource_dict = {
            self.resource_id: {
                "type": self._type,
                "properties": self._properties
            }
        }
        return resource_dict


class HeatTemplate(object):
    heat_template_version = str(datetime.date(2015, 10, 15))
    description = "karbor restore template"

    def __init__(self):
        super(HeatTemplate, self).__init__()
        self._resources = []
        self._original_id_resource_map = {}
        self._original_id_parameter_map = {}

    def put_resource(self, original_id, heat_resource):
        self._resources.append(heat_resource)
        self._original_id_resource_map[original_id] = heat_resource.resource_id

    def put_parameter(self, original_id, parameter):
        self._original_id_parameter_map[original_id] = parameter

    def get_resource_reference(self, original_id):
        if original_id in self._original_id_resource_map:
            return {
                "get_resource": (self._original_id_resource_map[original_id])
            }
        elif original_id in self._original_id_parameter_map:
            return self._original_id_parameter_map[original_id]
        else:
            LOG.error("The reference is not found, original_id:%s",
                      original_id)
            raise InvalidOriginalId

    def len(self):
        return len(self._resources)

    def to_dict(self):
        resources_dict = {}
        for resource in self._resources:
            resource_id = resource.resource_id
            resource_dict = resource.to_dict()
            resources_dict[resource_id] = resource_dict[resource_id]
        template_dict = {
            "heat_template_version": self.heat_template_version,
            "description": self.description,
            "resources": resources_dict
        }
        return yaml.load(jsonutils.dumps(template_dict))

    def dump_to_yaml_file(self, file_name):
        with open(file_name, "w") as f:
            yaml.dump(yaml.load(jsonutils.dumps(self.to_dict())),
                      f,
                      default_flow_style=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.resource import Resource
from karbor.services.protection import bank_plugin
from karbor.services.protection import checkpoint
from karbor.services.protection import graph

from karbor.tests import base
from karbor.tests.unit.protection.fakes import fake_protection_plan
from karbor.tests.unit.protection.test_bank import _InMemoryBankPlugin
from karbor.tests.unit.protection.test_bank import _InMemoryLeasePlugin

A = Resource(id="A", type="fake", name="fake")
B = Resource(id="B", type="fake", name="fake")
C = Resource(id="C", type="fake", name="fake")
D = Resource(id="D", type="fake", name="fake")
E = Resource(id="E", type="fake", name="fake")

resource_map = {
    A: [C],
    B: [C],
    C: [D, E],
    D: [],
    E: [],
}


class CheckpointTest(base.TestCase):
    def test_create_in_section(self):
        bank = bank_plugin.Bank(_InMemoryBankPlugin())
        bank_lease = _InMemoryLeasePlugin()
        checkpoints_section = bank_plugin.BankSection(bank, "/checkpoints")
        indices_section = bank_plugin.BankSection(bank, "/indices")
        owner_id = bank.get_owner_id()
        plan = fake_protection_plan()
        cp = checkpoint.Checkpoint.create_in_section(
            checkpoints_section=checkpoints_section,
            indices_section=indices_section,
            bank_lease=bank_lease,
            owner_id=owner_id,
            plan=plan)
        checkpoint_data = cp._md_cache
        self.assertEqual(
            checkpoint_data,
            bank._plugin.get_object(
                "/checkpoints/%s/%s" % (checkpoint_data['id'],
                                        checkpoint._INDEX_FILE_NAME)
            )
        )
        self.assertEqual(owner_id, cp.owner_id)
        self.assertEqual("protecting", cp.status)

    def test_resource_graph(self):
        bank = bank_plugin.Bank(_InMemoryBankPlugin())
        bank_lease = _InMemoryLeasePlugin()
        checkpoints_section = bank_plugin.BankSection(bank, "/checkpoints")
        indices_section = bank_plugin.BankSection(bank, "/indices")
        owner_id = bank.get_owner_id()
        plan = fake_protection_plan()
        cp = checkpoint.Checkpoint.create_in_section(
            checkpoints_section=checkpoints_section,
            indices_section=indices_section,
            bank_lease=bank_lease,
            owner_id=owner_id,
            plan=plan)

        resource_graph = graph.build_graph([A, B, C, D],
                                           resource_map.__getitem__)
        cp.resource_graph = resource_graph
        cp.commit()
        checkpoint_data = cp._md_cache
        self.assertEqual(
            checkpoint_data,
            bank._plugin.get_object(
                "/checkpoints/%s/%s" % (checkpoint_data["id"],
                                        checkpoint._INDEX_FILE_NAME)
            )
        )
        self.assertEqual(len(resource_graph), len(cp.resource_graph))
        for start_node in resource_graph:
            self.assertIn(start_node, cp.resource_graph)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import uuidutils

from karbor.common import constants
from karbor.services.protection import client_factory
from karbor.services.protection import resource_flow
from karbor.services.protection import restore_heat
from taskflow import task

sync_status_opts = [
    cfg.IntOpt('sync_status_interval',
               default=20,
               help='update protection status interval')
]

CONF = cfg.CONF
CONF.register_opts(sync_status_opts)

LOG = logging.getLogger(__name__)


class InitiateRestoreTask(task.Task):
    def execute(self, restore, *args, **kwargs):
        LOG.debug("Initiate restore restore_id: %s", restore.id)
        restore['status'] = constants.RESTORE_STATUS_IN_PROGRESS
        restore.save()

    def revert(self, restore, *args, **kwargs):
        LOG.debug("Failed to restore restore_id: %s", restore.id)
        restore['status'] = constants.RESTORE_STATUS_FAILURE
        restore.save()


class CompleteRestoreTask(task.Task):
    def execute(self, restore, *args, **kwargs):
        LOG.debug("Complete restore restore_id: %s", restore.id)
        restore['status'] = constants.RESTORE_STATUS_SUCCESS
        restore.save()


class CreateHeatTask(task.Task):
    default_provides = ['heat_client', 'heat_template']

    def execute(self, context, heat_conf):
        LOG.info('Creating Heat template. Target: "%s"'
                 % heat_conf.get('auth_url', '(None)'))
        heat_client = client_factory.ClientFactory.create_client(
            'heat', context=context, **heat_conf)

        heat_template = restore_heat.HeatTemplate()

        return (heat_client, heat_template)


class CreateStackTask(task.Task):
    default_provides = 'stack_id'

    def execute(self, heat_client, heat_template):
        stack_name = "restore_%s" % uuidutils.generate_uuid()
        if heat_template.len() == 0:
            LOG.info('Not creating Heat stack, no resources in template')
            return None
        LOG.info('Creating Heat stack, stack_name: %s', stack_name)
        try:
            body = heat_client.stacks.create(
                stack_name=stack_name,
                template=heat_template.to_dict())
            LOG.debug('Created stack with id: %s', body['stack']['id'])
            return body['stack']['id']
        except Exception:
            LOG.error("use heat to create stack failed")
            raise


class SyncRestoreStatusTask(task.Task):
    def execute(self, stack_id, heat_client, restore):
        if stack_id is None:
            LOG.info('Not syncing Heat stack status, stack is empty')
            return

        LOG.info('Syncing Heat stack status, stack_id: %s', stack_id)
        self._restore = restore
        sync_status_loop = loopingcall.FixedIntervalLoopingCall(
            self._sync_status, heat_client, stack_id)
        sync_status_loop.start(interval=CONF.sync_status_interval)
        sync_status_loop.wait()

    def _sync_status(self, heat_client, stack_id):
        try:
            stack = heat_client.stacks.get(stack_id)
        except Exception:
            LOG.debug('Heat error getting stack, stack_id: %s', stack_id)
            raise
        stack_status = getattr(stack, 'stack_status')
        if stack_status == 'CREATE_IN_PROGRESS':
            LOG.debug('Heat stack status: in progress, stack_id: %s',
                      stack_id)
        elif stack_status == 'CREATE_COMPLETE':
            LOG.info('Heat stack status: complete, stack_id: %s', stack_id)
            self._update_resource_status(heat_client, stack_id)
            raise loopingcall.LoopingCallDone()
        else:
            LOG.info('Heat stack status: failure, stack_id: %s', stack_id)
            self._update_resource_status(heat_client, stack_id)
            raise

    def _update_resource_status(self, heat_client, stack_id):
        LOG.debug('Updating resources status from heat stack (stack_id: %s)',
                  stack_id)
        try:
            resources = heat_client.resources.list(stack_id)
            for resource in resources:
                heat_to_karbor_map = {
                    'CREATE_COMPLETE': constants.RESOURCE_STATUS_AVAILABLE,
                    'CREATE_IN_PROGRESS': constants.RESOURCE_STATUS_RESTORING,
                    'CREATE_FAILED': constants.RESOURCE_STATUS_ERROR,
                }
                reason = resource.resource_status_reason if (
                    resource.resource_status == 'CREATE_FAILED'
                ) else None
                self._restore.update_resource_status(
                    resource.resource_type,
                    resource.physical_resource_id,
                    heat_to_karbor_map[resource.resource_status],
                    reason,
                )
            self._restore.save()
        except Exception as e:
            LOG.warning('Unable to update resources status from heat stack. '
                        'Reason: %s', e)


def get_flow(context, workflow_engine, checkpoint, provider, restore,
             restore_auth):
    target = restore.get('restore_target', None)

    heat_conf = {}
    if target is not None:
        heat_conf["auth_url"] = target
        if restore_auth is not None:
            auth_type = restore_auth.get("type", None)
            if auth_type == "password":
                heat_conf["username"] = restore_auth["username"]
                heat_conf["password"] = restore_auth["password"]

    resource_graph = checkpoint.resource_graph
    parameters = restore.parameters
    flow_name = "Restore_" + checkpoint.id
    restore_flow = workflow_engine.build_flow(flow_name, 'linear')
    plugins = provider.load_plugins()
    resources_task_flow = resource_flow.build_resource_flow(
        operation_type=constants.OPERATION_RESTORE,
        context=context,
        workflow_engine=workflow_engine,
        resource_graph=resource_graph,
        plugins=plugins,
        parameters=parameters
    )

    workflow_engine.add_tasks(
        restore_flow,
        InitiateRestoreTask(),
        CreateHeatTask(inject={'context': context, 'heat_conf': heat_conf}),
        resources_task_flow,
        CreateStackTask(),
        SyncRestoreStatusTask(),
        CompleteRestoreTask()
    )
    flow_engine = workflow_engine.get_engine(restore_flow,
                                             store={'checkpoint': checkpoint,
                                                    'restore': restore})
    return flow_engine



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The triggers api."""

from datetime import datetime
from oslo_log import log as logging
from oslo_utils import uuidutils
from webob import exc

from karbor.api import common
from karbor.api.openstack import wsgi
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor import policy
from karbor.services.operationengine import api as operationengine_api
from karbor import utils

LOG = logging.getLogger(__name__)


def check_policy(context, action, target_obj=None):
    _action = 'trigger:%s' % action
    policy.enforce(context, _action, target_obj)


class TriggerViewBuilder(common.ViewBuilder):
    """Model a trigger API response as a python dictionary."""

    _collection_name = "triggers"

    def detail(self, request, trigger):
        """Detailed view of a single trigger."""

        trigger_ref = {
            'trigger_info': {
                'id': trigger.get('id'),
                'name': trigger.get('name'),
                'type': trigger.get('type'),
                'properties': trigger.get('properties'),
            }
        }
        return trigger_ref

    def detail_list(self, request, triggers):
        """Detailed view of a list of triggers."""
        return self._list_view(self.detail, request, triggers)

    def _list_view(self, func, request, triggers):
        triggers_list = [func(request, item)['trigger_info']
                         for item in triggers]

        triggers_links = self._get_collection_links(request,
                                                    triggers,
                                                    self._collection_name,
                                                    )
        ret = {'triggers': triggers_list}
        if triggers_links:
            ret['triggers_links'] = triggers_links

        return ret


class TriggersController(wsgi.Controller):
    """The Triggers API controller for the OpenStack API."""

    _view_builder_class = TriggerViewBuilder

    def __init__(self):
        self.operationengine_api = operationengine_api.API()
        super(TriggersController, self).__init__()

    def create(self, req, body):
        """Creates a new trigger."""

        LOG.debug('Create trigger start')

        if not self.is_valid_body(body, 'trigger_info'):
            raise exc.HTTPUnprocessableEntity()
        LOG.debug('Create a trigger, request body: %s', body)

        context = req.environ['karbor.context']
        check_policy(context, 'create')
        trigger_info = body['trigger_info']

        trigger_name = trigger_info.get("name", None)
        trigger_type = trigger_info.get("type", None)
        trigger_property = trigger_info.get("properties", None)
        if not trigger_name or not trigger_type or not trigger_property:
            msg = _("Trigger name or type or property is not provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        self.validate_name_and_description(trigger_info)

        trigger_property.setdefault(
            'start_time', datetime.utcnow().replace(microsecond=0))
        trigger_definition = {
            'id': uuidutils.generate_uuid(),
            'name': trigger_name,
            'project_id': context.project_id,
            'type': trigger_type,
            'properties': trigger_property,
        }
        try:
            trigger = objects.Trigger(context=context, **trigger_definition)
            self.operationengine_api.create_trigger(context, trigger)
            trigger.create()
        except exception.Invalid as ex:
            raise exc.HTTPBadRequest(explanation=ex.msg)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return self._view_builder.detail(req, trigger)

    def delete(self, req, id):
        """Delete a trigger."""

        LOG.debug('Delete trigger(%s) start', id)

        context = req.environ['karbor.context']
        trigger = self._get_trigger_by_id(context, id)

        check_policy(context, 'delete', trigger)

        try:
            operations = objects.ScheduledOperationList.get_by_filters(
                context, {"trigger_id": id}, limit=1)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        if operations:
            msg = _("Trigger is being used by one or more operations")
            raise exc.HTTPFailedDependency(explanation=msg)

        try:
            self.operationengine_api.delete_trigger(context, id)
        except exception.TriggerNotFound as ex:
            pass
        except (exception.DeleteTriggerNotAllowed,
                Exception) as ex:
            self._raise_unknown_exception(ex)

        trigger.destroy()

    def update(self, req, id, body):
        """Update a trigger"""

        LOG.debug('Update trigger(%s) start', id)

        context = req.environ['karbor.context']
        trigger = self._get_trigger_by_id(context, id)

        check_policy(context, 'update', trigger)

        trigger_info = body['trigger_info']
        trigger_name = trigger_info.get("name", None)
        trigger_property = trigger_info.get("properties", None)

        if trigger_name:
            self.validate_name_and_description(trigger_info)
            trigger.name = trigger_name

        if trigger_property:
            start_time = trigger_property.get('start_time', None)
            if not start_time:
                msg = (_("start_time should be supplied"))
                raise exc.HTTPBadRequest(explanation=msg)
            try:
                trigger.properties = trigger_property
                self.operationengine_api.update_trigger(context, trigger)
            except exception.InvalidInput as ex:
                raise exc.HTTPBadRequest(explanation=ex.msg)
            except (exception.TriggerNotFound, Exception) as ex:
                self._raise_unknown_exception(ex)
        try:
            trigger.save()
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return self._view_builder.detail(req, trigger)

    def show(self, req, id):
        """Return data about the given trigger."""

        LOG.debug('Get trigger(%s) start', id)

        context = req.environ['karbor.context']
        trigger = self._get_trigger_by_id(context, id)

        check_policy(context, 'get', trigger)
        return self._view_builder.detail(req, trigger)

    def index(self, req):
        """Returns a list of triggers, transformed through view builder."""

        context = req.environ['karbor.context']
        check_policy(context, 'list')

        params = req.params.copy()
        LOG.debug('List triggers start, params=%s', params)
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        valid_filters = ["all_tenants", "name", "type", "properties"]
        utils.remove_invalid_filter_options(context, filters, valid_filters)
        utils.check_filters(filters)

        all_tenants = utils.get_bool_param("all_tenants", filters)
        if not (context.is_admin and all_tenants):
            filters["project_id"] = context.project_id

        try:
            triggers = objects.TriggerList.get_by_filters(
                context, filters, limit, marker, sort_keys, sort_dirs)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return self._view_builder.detail_list(req, triggers)

    def _get_trigger_by_id(self, context, id):
        if not uuidutils.is_uuid_like(id):
            msg = _("Invalid trigger id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            trigger = objects.Trigger.get_by_id(context, id)
        except exception.TriggerNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)
        except Exception as ex:
            self._raise_unknown_exception(ex)

        return trigger

    def _raise_unknown_exception(self, exception_instance):
        LOG.exception('An unknown exception happened')

        value = exception_instance.msg if isinstance(
            exception_instance, exception.KarborException) else type(
            exception_instance)
        msg = (_('Unexpected API Error. Please report this at '
                 'http://bugs.launchpad.net/karbor/ and attach the '
                 'Karbor API log if possible.\n%s') % value)
        raise exc.HTTPInternalServerError(explanation=msg)


def create_resource():
    return wsgi.Resource(TriggersController())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the OperationEngine manager RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging

from karbor.objects import base as objects_base
from karbor import rpc


CONF = cfg.CONF


class OperationEngineAPI(object):
    """Client side of the OperationEngine rpc API.

    API version history:

        1.0 - Initial version.
    """

    RPC_API_VERSION = '1.0'

    def __init__(self):
        super(OperationEngineAPI, self).__init__()
        target = messaging.Target(topic=CONF.operationengine_topic,
                                  version=self.RPC_API_VERSION)
        serializer = objects_base.KarborObjectSerializer()

        client = rpc.get_client(target, version_cap=None,
                                serializer=serializer)
        self._client = client.prepare(version='1.0')

    def create_scheduled_operation(self, ctxt, operation):
        return self._client.call(ctxt, 'create_scheduled_operation',
                                 operation=operation)

    def delete_scheduled_operation(self, ctxt, operation_id, trigger_id):
        return self._client.call(ctxt, 'delete_scheduled_operation',
                                 operation_id=operation_id,
                                 trigger_id=trigger_id)

    def suspend_scheduled_operation(self, ctxt, operation_id, trigger_id):
        return self._client.call(ctxt, 'suspend_scheduled_operation',
                                 operation_id=operation_id,
                                 trigger_id=trigger_id)

    def resume_scheduled_operation(self, ctxt, operation_id, trigger_id):
        return self._client.call(ctxt, 'resume_scheduled_operation',
                                 operation_id=operation_id,
                                 trigger_id=trigger_id)

    def create_trigger(self, ctxt, trigger):
        return self._client.call(ctxt, 'create_trigger', trigger=trigger)

    def delete_trigger(self, ctxt, trigger_id):
        return self._client.call(ctxt, 'delete_trigger', trigger_id=trigger_id)

    def update_trigger(self, ctxt, trigger):
        return self._client.call(ctxt, 'update_trigger', trigger=trigger)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Resource(object):
    __slots__ = ('type', 'id', 'name', 'extra_info')

    def __init__(self, type, id, name, extra_info=None):
        self.type = type
        self.id = id
        self.name = name
        self.extra_info = extra_info

    def __setattr__(self, key, value):
        try:
            getattr(self, key)
        except AttributeError:
            pass
        else:
            raise AttributeError()

        return super(Resource, self).__setattr__(key, value)

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key

    def to_dict(self):
        return {item: getattr(self, item) for item in self.__slots__}

    @property
    def key(self):
        return (self.type, self.id, self.name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

OPTIONS_SCHEMA = {
    "title": "Cinder Protection Options",
    "type": "object",
    "properties": {
        "backup_name": {
            "type": "string",
            "title": "Backup Name",
            "description": "The name of the backup.",
            "default": None
        },
        "backup_mode": {
            "type": "string",
            "title": "Backup Mode",
            "description": "The backup mode.",
            "enum": ["full", "incremental"],
            "default": "full"
        },
        "container": {
            "type": "string",
            "title": "Container",
            "description": "The container which been chosen.",
            "default": None
        },
        "description": {
            "type": "string",
            "title": "Description",
            "description": "The description of the volume.",
            "default": None
        },
        "force": {
            "type": "boolean",
            "title": "Force",
            "description": "Whether to backup, even if the volume "
                           "is attached",
            "default": False
        }
    },
    "required": ["backup_name", "backup_mode", "container", "force"]
}

RESTORE_SCHEMA = {
    "title": "Cinder Protection Restore",
    "type": "object",
    "properties": {
        "volume_id": {
            "type": "string",
            "title": "Volume ID",
            "description": "The target volume ID to restore to.",
            "default": None
        },
        "restore_name": {
            "type": "string",
            "title": "Restore Name",
            "description": "The name of the restored volume.",
            "default": None
        },
        "restore_description": {
            "type": "string",
            "title": "Restore Description",
            "description": "The description of the restored volume.",
            "default": None
        }
    }
}

SAVED_INFO_SCHEMA = {
    "title": "Cinder Protection Saved Info",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "title": "Name",
            "description": "The name for this backup."
        },
        "is_incremental": {
            "type": "boolean",
            "title": "Is Incremental",
            "description":
                "The type of the backup, "
                "True is incremental and False is full."
        },
        "status": {
            "type": "string",
            "title": "Status",
            "description": "The backup status, such as available.",
            "enum": ['creating', 'available',
                                 'deleting', 'error',
                                 'restoring', 'error_restoring'],
        },
        "progress": {
            "type": "number",
            "title": "Progress",
            "description":
            "The current operation progress for this backup.",
            "constraint": {'min': 0, 'max': 1},
        },
        "fail_reason": {
            "type": "string",
            "title": "Fail Reason",
            "description":
                "The reason for the failure status of the backup."
        },
        "size": {
            "type": "integer",
            "title": "Size",
            "description": "The size of the backup, in GB."
        },
        "volume_id": {
            "type": "string",
            "title": "Volume ID",
            "description":
                ("The ID of the volume "
                 "from which the backup was created.")
        },
    },
    "required": ["name", "status", "progress", "fail_reason",
                 "size", "volume_id"]
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Manage all triggers.
"""

from karbor import exception
from karbor.i18n import _
from karbor.services.operationengine.engine import triggers as all_triggers


class TriggerManager(object):
    """Manage all trigger classes which are defined at triggers dir."""

    def __init__(self, executor):
        super(TriggerManager, self).__init__()

        all_cls = all_triggers.all_triggers()
        self._trigger_cls_map = {cls.TRIGGER_TYPE:
                                 cls for cls in all_cls}

        for t, cls in self._trigger_cls_map.items():
            cls.check_configuration()

        # self._trigger_obj_map = {
        #     trigger_id: trigger,
        # }
        self._trigger_obj_map = {}

        self._executor = executor

    def shutdown(self):

        for trigger_id, trigger in self._trigger_obj_map.items():
            trigger.shutdown()

        self._trigger_obj_map.clear()
        self._trigger_cls_map.clear()

        if self._executor:
            self._executor.shutdown()
            self._executor = None

    def check_trigger_definition(self, trigger_type, trigger_definition):
        """Check trigger definition

        :param trigger_type: Type of trigger
        :param trigger_definition: Definition of trigger
        """

        trigger_cls = self._get_trigger_class(trigger_type)
        trigger_cls.check_trigger_definition(trigger_definition)

    def add_trigger(self, trigger_id, trigger_type, trigger_property):
        if trigger_id in self._trigger_obj_map:
            msg = (_("Trigger id(%s) is exist") % trigger_id)
            raise exception.InvalidInput(msg)

        trigger_cls = self._get_trigger_class(trigger_type)
        trigger = trigger_cls(trigger_id, trigger_property, self._executor)
        self._trigger_obj_map[trigger_id] = trigger

    def remove_trigger(self, trigger_id):
        trigger = self._trigger_obj_map.get(trigger_id, None)
        if not trigger:
            raise exception.TriggerNotFound(id=trigger_id)

        if trigger.has_operations():
            raise exception.DeleteTriggerNotAllowed(trigger_id=trigger_id)

        trigger.shutdown()
        del self._trigger_obj_map[trigger_id]

    def update_trigger(self, trigger_id, trigger_property):
        trigger = self._trigger_obj_map.get(trigger_id, None)
        if not trigger:
            raise exception.TriggerNotFound(id=trigger_id)

        trigger.update_trigger_property(trigger_property)

    def register_operation(self, trigger_id, operation_id, **kwargs):
        """Register operation definition.

        :param trigger_id: The ID of the trigger which
                           the operation is registered to
        :param operation_id: ID of the operation
        :param kwargs: Any parameters
        :raise InvalidInput if the trigger_type is invalid or
               other exceptionis register_operation of trigger raises
        """
        trigger = self._trigger_obj_map.get(trigger_id, None)
        if not trigger:
            raise exception.TriggerNotFound(id=trigger_id)

        trigger.register_operation(operation_id, **kwargs)

        if kwargs.get('resume'):
            self._executor.resume_operation(operation_id, **kwargs)

    def unregister_operation(self, trigger_id, operation_id, **kwargs):
        """Unregister operation.

        :param trigger_id: The ID of the trigger which
                           the operation is registered to
        :param operation_id: ID of the operation
        :raise InvalidInput if the trigger_type is invalid or
               other exceptionis unregister_operation of trigger raises
        """
        trigger = self._trigger_obj_map.get(trigger_id, None)
        if not trigger:
            raise exception.TriggerNotFound(id=trigger_id)

        trigger.unregister_operation(operation_id, **kwargs)
        self._executor.cancel_operation(operation_id)

    def _get_trigger_class(self, trigger_type):
        cls = self._trigger_cls_map.get(trigger_type, None)
        if not cls:
            msg = (_("Invalid trigger type:%s") % trigger_type)
            raise exception.InvalidInput(msg)

        return cls



#!/usr/bin/env python
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Starter script for karbor protection service."""

import eventlet
eventlet.monkey_patch()

import sys

from oslo_config import cfg
from oslo_log import log as logging

from karbor import i18n
i18n.enable_lazy()
from karbor import objects

# Need to register global_opts
from karbor.common import config  # noqa
from karbor import service
from karbor import version


CONF = cfg.CONF


def main():
    objects.register_all()
    CONF(sys.argv[1:], project='karbor',
         version=version.version_string())
    logging.setup(CONF, "karbor")
    server = service.Service.create(binary='karbor-protection')
    service.serve(server)
    service.wait()
